# Copyright 2016 - Nokia, ZTE
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from testtools import matchers

from vitrage.common.constants import DatasourceAction
from vitrage.common.constants import DatasourceOpts as DSOpts
from vitrage.common.constants import GraphAction
from vitrage.datasources.static import driver
from vitrage.datasources.static import STATIC_DATASOURCE
from vitrage.datasources.static import StaticFields
from vitrage.tests import base
from vitrage.tests.mocks import utils


class TestStaticDriver(base.BaseTest):

    CHANGES_DIR = '/changes_datasources'

    # noinspection PyAttributeOutsideInit,PyPep8Naming
    @classmethod
    def setUpClass(cls):
        super(TestStaticDriver, cls).setUpClass()
        cls.static_driver = driver.StaticDriver()

    def test_get_all(self):
        self._set_conf()

        # Action
        static_entities = self.static_driver.get_all(
            DatasourceAction.INIT_SNAPSHOT)

        # Test assertions
        self.assertThat(static_entities, matchers.HasLength(8))

        for entity in static_entities:
            self._validate_static_entity(entity)

    # noinspection PyAttributeOutsideInit
    def test_get_changes_with_added_resources(self):
        # Get initial resources
        self._set_conf(self.CHANGES_DIR + '/baseline')

        entities = self.static_driver.get_all(DatasourceAction.UPDATE)
        self.assertThat(entities, matchers.HasLength(4))

        # Add resources
        self._set_conf(self.CHANGES_DIR + '/added_resources')

        # Action
        changes = self.static_driver.get_changes(
            GraphAction.UPDATE_ENTITY)

        # Test Assertions
        expected_changes = [
            {'static_id': 's3', 'type': 'switch', 'name': 'switch-3 is new!',
             'id': '3333', 'state': 'available'},
            {'static_id': 'r2', 'type': 'router', 'name': 'router-2 is new!',
             'id': '2222', 'state': 'available'},
        ]

        self._validate_static_changes(expected_changes, changes)

    # noinspection PyAttributeOutsideInit
    def test_get_changes_with_deleted_resources(self):
        # Get initial resources
        self._set_conf(self.CHANGES_DIR + '/baseline')

        entities = self.static_driver.get_all(DatasourceAction.UPDATE)
        self.assertThat(entities, matchers.HasLength(4))

        # Delete resources
        self._set_conf(self.CHANGES_DIR + '/deleted_resources')

        # Action
        changes = self.static_driver.get_changes(
            GraphAction.UPDATE_ENTITY)

        # Test Assertions
        expected_changes = [
            {'static_id': 's2', 'type': 'switch', 'name': 'switch-2',
             'id': '23456', 'state': 'available',
             'vitrage_event_type': 'delete_entity'},
            {'static_id': 'h1', 'type': 'nova.host', 'id': '1',
             'vitrage_event_type': 'delete_entity'},
        ]

        self._validate_static_changes(expected_changes, changes)

    # noinspection PyAttributeOutsideInit
    def test_get_changes_with_changed_resources(self):
        # Get initial resources
        self._set_conf(self.CHANGES_DIR + '/baseline')

        entities = self.static_driver.get_all(DatasourceAction.UPDATE)
        self.assertThat(entities, matchers.HasLength(4))

        # Delete resources
        self._set_conf(self.CHANGES_DIR + '/changed_resources')

        # Action
        changes = self.static_driver.get_changes(
            GraphAction.UPDATE_ENTITY)

        # Test Assertions
        expected_changes = [
            {'static_id': 's1', 'type': 'switch', 'name': 'switch-1',
             'id': '12345', 'state': 'error'},
            {'static_id': 'r1', 'type': 'router',
             'name': 'router-1 is the best!', 'id': '45678'},
        ]

        self._validate_static_changes(expected_changes, changes)

    # noinspection PyAttributeOutsideInit
    def test_get_changes_with_mixed_changes(self):
        # Get initial resources
        self._set_conf(self.CHANGES_DIR + '/baseline')

        entities = self.static_driver.get_all(DatasourceAction.UPDATE)
        self.assertThat(entities, matchers.HasLength(4))

        # Delete resources
        self._set_conf(self.CHANGES_DIR + '/mixed_changes')

        # Action
        changes = self.static_driver.get_changes(
            GraphAction.UPDATE_ENTITY)

        # Test Assertions
        expected_changes = [
            {'static_id': 's1', 'type': 'switch', 'name': 'switch-1',
             'id': '12345', 'state': 'available',
             'vitrage_event_type': 'delete_entity'},
            {'static_id': 's2', 'type': 'switch', 'name': 'switch-2',
             'id': '23456', 'state': 'error'},
            {'static_id': 'r2', 'type': 'router', 'name': 'router-2 is new!',
             'id': '222'},
        ]

        self._validate_static_changes(expected_changes, changes)

    def _validate_static_entity(self, entity):
        self.assertIsInstance(entity[StaticFields.METADATA], dict)
        for rel in entity[StaticFields.RELATIONSHIPS]:
            self._validate_static_rel(entity, rel)

    def _validate_static_rel(self, entity, rel):
        self.assertTrue(entity[StaticFields.STATIC_ID] in
                        (rel[StaticFields.SOURCE], rel[StaticFields.TARGET]))
        self.assertTrue(
            isinstance(rel[StaticFields.SOURCE], dict)
            and entity[StaticFields.STATIC_ID] == rel[StaticFields.TARGET]
            or isinstance(rel[StaticFields.TARGET], dict)
            and entity[StaticFields.STATIC_ID] == rel[StaticFields.SOURCE]
            or entity[StaticFields.STATIC_ID] == rel[StaticFields.SOURCE]
            and entity[StaticFields.STATIC_ID] == rel[StaticFields.TARGET])

    def _validate_static_changes(self, expected_changes, changes):
        self.assertThat(changes, matchers.HasLength(len(expected_changes)))

        for entity in changes:
            self._validate_static_entity(entity)

        for expected_change in expected_changes:
            found = False
            for change in changes:
                if change[StaticFields.TYPE] == \
                        expected_change[StaticFields.TYPE] and \
                        change[StaticFields.ID] == \
                        expected_change[StaticFields.ID]:
                    found = True
                    self.assertEqual(expected_change.get('vitrage_event_type'),
                                     change.get('vitrage_event_type'))
                    self.assertEqual(expected_change.get(StaticFields.NAME),
                                     change.get(StaticFields.NAME))
                    self.assertEqual(expected_change.get(StaticFields.STATE),
                                     change.get(StaticFields.STATE))
            self.assertTrue(found)

    def _set_conf(self, sub_dir=None):
        default_dir = utils.get_resources_dir() + \
            '/static_datasources' + (sub_dir if sub_dir else '')

        opts = [
            cfg.StrOpt(DSOpts.TRANSFORMER,
                       default='vitrage.datasources.static.transformer.'
                               'StaticTransformer'),
            cfg.StrOpt(DSOpts.DRIVER,
                       default='vitrage.datasources.static.driver.'
                               'StaticDriver'),
            cfg.IntOpt(DSOpts.CHANGES_INTERVAL,
                       default=30,
                       min=30,
                       help='interval between checking changes in the static '
                            'datasources'),
            cfg.StrOpt('directory', default=default_dir),
        ]

        self.conf_reregister_opts(opts, group=STATIC_DATASOURCE)



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from vitrage.common.constants import VertexProperties as VProps
from vitrage.datasources.alarm_properties import AlarmProperties as AlarmProps
from vitrage.evaluator.actions.recipes.action_steps import ADD_VERTEX
from vitrage.evaluator.actions.recipes.action_steps import REMOVE_VERTEX
from vitrage.evaluator.actions.recipes import base
from vitrage.evaluator.actions.recipes.base import ActionStepWrapper
from vitrage.evaluator.template_fields import TemplateFields as TFields


class RaiseAlarm(base.Recipe):
    """Raise a Vitrage (deduced) alarm.

    The 'get_do_recipe' and 'get_undo_recipe' receive action_spec as input.
    The action_spec contains the following fields: type, targets and
    properties. example input:

    action_spec = ActionSpecs('type'= {'raise_alarm'},
                              'targets'= {target: vertex},
                              'properties' = {severity : CRITICAL,
                                              alarm_name: instance_cpu_problem}
    """

    @staticmethod
    def get_do_recipe(action_spec):

        params = RaiseAlarm._get_vertex_params(action_spec)
        params[VProps.STATE] = AlarmProps.ACTIVE_STATE
        add_vertex_step = ActionStepWrapper(ADD_VERTEX, params)

        return [add_vertex_step]

    @staticmethod
    def get_undo_recipe(action_spec):

        params = RaiseAlarm._get_vertex_params(action_spec)
        params[VProps.STATE] = AlarmProps.INACTIVE_STATE
        remove_vertex_step = ActionStepWrapper(REMOVE_VERTEX, params)

        return [remove_vertex_step]

    @staticmethod
    def _get_vertex_params(action_spec):

        target_resource = action_spec.targets[TFields.TARGET]
        add_vertex_params = {
            TFields.TARGET: target_resource.vertex_id,
            VProps.VITRAGE_RESOURCE_TYPE: target_resource[VProps.VITRAGE_TYPE],
        }
        add_vertex_params.update(action_spec.properties)

        return add_vertex_params



#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_config import cfg
from oslo_db import options as db_options
from oslo_log import log
from oslo_policy import opts as policy_opts
from osprofiler import initializer as osprofiler_initializer
from osprofiler import opts as osprofiler_opts
from vitrage import keystone_client
from vitrage import messaging
from vitrage import opts


CONF = cfg.CONF
LOG = log.getLogger(__name__)


def parse_config(args, default_config_files=None):
    set_defaults()
    log.register_options(CONF)
    policy_opts.set_defaults(CONF)
    osprofiler_opts.set_defaults(CONF)
    db_options.set_defaults(CONF)

    for group, options in opts.list_opts():
        CONF.register_opts(list(options),
                           group=None if group == 'DEFAULT' else group)

    CONF(args[1:], project='vitrage', validate_default_values=True,
         default_config_files=default_config_files)

    if CONF.profiler.enabled:
        osprofiler_initializer.init_from_conf(
            conf=CONF,
            context=None,
            project='vitrage',
            service='api',
            host=CONF.api.host
        )

    for datasource in CONF.datasources.types:
        opts.register_opts(datasource, CONF.datasources.path)

    keystone_client.register_keystoneauth_opts()
    log.setup(CONF, 'vitrage')
    CONF.log_opt_values(LOG, log.DEBUG)
    messaging.setup()


def set_defaults():
    from oslo_middleware import cors
    cfg.set_defaults(cors.CORS_OPTS,
                     allow_headers=[
                         'Authorization',
                         'X-Auth-Token',
                         'X-Subject-Token',
                         'X-User-Id',
                         'X-Domain-Id',
                         'X-Project-Id',
                         'X-Roles'])



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from vitrage.common.constants import VertexProperties as VProps
from vitrage.entity_graph.mappings.handler_base import HandlerBase
from vitrage.entity_graph.mappings.operational_alarm_severity \
    import OperationalAlarmSeverity


class AlarmHandler(HandlerBase):

    def __init__(self):
        super(AlarmHandler, self).__init__()

    def undefined_property(self):
        return OperationalAlarmSeverity.NA

    def value_properties(self):
        return [VProps.SEVERITY]

    def set_operational_value(self, new_vertex, operational_value):
        new_vertex[VProps.VITRAGE_OPERATIONAL_SEVERITY] = operational_value

    def set_aggregated_value(self, new_vertex, aggregated_value):
        new_vertex[VProps.VITRAGE_AGGREGATED_SEVERITY] = aggregated_value

    def default_values(self):
        return [(None, OperationalAlarmSeverity.NA, 0)]

    def get_value_class_instance(self):
        return OperationalAlarmSeverity()



# -*- coding: utf-8 -*-
#
# © 2013 Lyft, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License.  You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Common code for collectd python plugins.
"""

import collectd


class CollectDPlugin(object):
    """Base class for collectd plugins written in Python.

    Each plugin must have a unique name which must match the name
    used for the configuration block in the collectd configuration file.
    """

    def __init__(self, name=None):
        self.name = name
        if name:
            collectd.register_config(self.configure, name=self.name)
        else:
            collectd.register_config(self.configure)
        collectd.register_init(self.initialize)
        collectd.register_shutdown(self.shutdown)

    @staticmethod
    def config_to_dict(config):
        """Convert a collectd.Config object to a dictionary. """

        def config_to_tuple(config):
            """Convert a collectd.Config object to a tuple. """

            if config.children:
                return (config.key, dict(config_to_tuple(child)
                                         for child in config.children))
            elif len(config.values) > 1:
                return config.key, config.values
            else:
                return config.key, config.values[0]

        return {k: v for k, v in [config_to_tuple(config)]}

    def error(self, message):
        """Log an error message to the collectd logger. """

        collectd.error('%s plugin: %s' % (self.name, message))

    def warning(self, message):
        """Log an warning message to the collectd logger. """

        collectd.warning('%s plugin: %s' % (self.name, message))

    def notice(self, message):
        """Log an notice message to the collectd logger. """

        collectd.notice('%s plugin: %s' % (self.name, message))

    def info(self, message):
        """Log an info message to the collectd logger. """

        collectd.info('%s plugin: %s' % (self.name, message))

    def debug(self, message):
        """Log an debug message to the collectd logger. """

        collectd.debug('%s plugin: %s' % (self.name, message))

    def configure(self, config, **kwargs):
        """Configuration callback for the plugin.

        will be called by collectd with a collectd.
        Config object containing configuration data for this plugin from the
        collectd configuration file.
        """

        # The top level of the configuration is the 'Module' block, which
        # is entirely useless, so we set the config attribute to its value,
        # which should be the interesting stuff.
        self.config = CollectDPlugin.config_to_dict(config)['Module']

    def initialize(self):
        """Initialization callback for the plugin.

        will be called by collectd with no arguments.
        """
        pass

    def shutdown(self):
        """Shutdown callback for the plugin.

        will be called by collectd with no arguments.
        """
        pass

    def add_read_callback(self, callback, **kwargs):
        """Register a read callback with collectd.

        kwargs will be passed to collectd.register_read.
        The callback will be called by collectd without arguments.
        """
        collectd.register_read(callback, **kwargs)

    def add_write_callback(self, callback, **kwargs):
        """Register a write callback with collectd.

        kwargs will be passed to collectd.register_read.
        The callback will be called by collectd with a collectd.
        Values object as the only argument.
        """

        collectd.register_write(callback)

    def add_flush_callback(self, callback, **kwargs):
        """Register a flush callback with collectd.

        kwargs will be passed to collectd.register_flush.
        The callback will be called by collectd with two arguments,
        a timeout and an identifier.
        """
        collectd.register_flush(callback, **kwargs)

    def add_log_callback(self, callback, **kwargs):
        """Register a log callback with collectd.

        kwargs will be passed to collectd.register_log.
        The callback will be called by collectd with two arguments,
        the severity and the message (without a newline at the end)
        """
        collectd.register_log(callback, **kwargs)

    def add_notification_callback(self, callback, **kwargs):
        """Register a notification callback with collectd.

        kwargs will be passed to collectd.register_notification.
        The callback will be called by collectd with a collectd.
        Notification object as the only argument.
        """
        collectd.register_notification(callback, **kwargs)


class PluginError(Exception):
    pass



# Copyright 2015 - Alcatel-Lucent
# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log
import sys

from vitrage.cli import VITRAGE_TITLE
from vitrage.common import config
from vitrage.common.utils import spawn
from vitrage.entity_graph.graph_init import VitrageGraphInit
from vitrage.entity_graph.workers import GraphWorkersManager

LOG = log.getLogger(__name__)


def main():
    """Main method of vitrage-graph"""

    print(VITRAGE_TITLE)
    config.parse_config(sys.argv)
    workers = GraphWorkersManager()
    spawn(init, workers)
    workers.run()


def init(workers):
    # Because fork duplicates the process memory.
    # We should only create master process resources after workers are forked.
    workers.wait_for_worker_start()
    VitrageGraphInit(workers).run()


if __name__ == "__main__":
    sys.exit(main())



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from vitrage.common.constants import DatasourceAction
from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.common.constants import GraphAction
from vitrage.common.exception import VitrageTransformerError
from vitrage.datasources.alarm_properties import AlarmProperties as AlarmProps
from vitrage.datasources import transformer_base as tbase


class AlarmTransformerBase(tbase.TransformerBase):

    def _ok_status(self, entity_event):
        pass

    def _extract_graph_action(self, entity_event):

        if DSProps.EVENT_TYPE in entity_event and \
           entity_event[DSProps.EVENT_TYPE] == GraphAction.DELETE_ENTITY:
            return entity_event[DSProps.EVENT_TYPE]

        datasource_action = entity_event[DSProps.DATASOURCE_ACTION]

        if datasource_action in \
            (DatasourceAction.UPDATE, DatasourceAction.SNAPSHOT):
            return GraphAction.DELETE_ENTITY \
                if self._ok_status(entity_event) else \
                self.GRAPH_ACTION_MAPPING.get(
                    entity_event.get(DSProps.EVENT_TYPE, None),
                    GraphAction.UPDATE_ENTITY)

        if DatasourceAction.INIT_SNAPSHOT == datasource_action:
            return GraphAction.CREATE_ENTITY

        raise VitrageTransformerError('Invalid datasource action: (%s)'
                                      % datasource_action)

    def _get_alarm_state(self, entity_event):
        event_type = entity_event.get(DSProps.EVENT_TYPE, None)
        if event_type is not None:
            return AlarmProps.INACTIVE_STATE if \
                GraphAction.DELETE_ENTITY == event_type else \
                AlarmProps.ACTIVE_STATE
        else:
            return AlarmProps.INACTIVE_STATE if \
                self._ok_status(entity_event) else \
                AlarmProps.ACTIVE_STATE



# Copyright 2016 - Alcatel-Lucent
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import datetime

from oslo_config import cfg
from oslo_log import log as logging
from testtools import matchers

from vitrage.common.constants import DatasourceAction
from vitrage.common.constants import DatasourceOpts as DSOpts
from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.common.constants import EdgeLabel
from vitrage.common.constants import EntityCategory
from vitrage.common.constants import GraphAction
from vitrage.common.constants import UpdateMethod
from vitrage.common.constants import VertexProperties as VProps
from vitrage.datasources.nova.host import NOVA_HOST_DATASOURCE
from vitrage.datasources.nova.host.transformer import HostTransformer
from vitrage.datasources.nova.zone import NOVA_ZONE_DATASOURCE
from vitrage.datasources.nova.zone.transformer import ZoneTransformer
from vitrage.datasources import transformer_base as tbase
from vitrage.datasources.transformer_base import TransformerBase
from vitrage.tests import base
from vitrage.tests.mocks import mock_driver as mock_sync

LOG = logging.getLogger(__name__)


# noinspection PyProtectedMember
class NovaHostTransformerTest(base.BaseTest):

    OPTS = [
        cfg.StrOpt(DSOpts.UPDATE_METHOD,
                   default=UpdateMethod.PUSH),
    ]

    # noinspection PyAttributeOutsideInit,PyPep8Naming
    @classmethod
    def setUpClass(cls):
        super(NovaHostTransformerTest, cls).setUpClass()
        cls.transformers = {}
        cls.conf = cfg.ConfigOpts()
        cls.conf.register_opts(cls.OPTS, group=NOVA_HOST_DATASOURCE)
        cls.transformers[NOVA_ZONE_DATASOURCE] = ZoneTransformer(
            cls.transformers)
        cls.transformers[NOVA_HOST_DATASOURCE] = HostTransformer(
            cls.transformers)

    def test_create_placeholder_vertex(self):
        LOG.debug('Nova host transformer test: Test create placeholder vertex')

        # Test setup
        host_name = 'host123'
        timestamp = datetime.datetime.utcnow()
        host_transformer = self.transformers[NOVA_HOST_DATASOURCE]

        # Test action
        properties = {
            VProps.ID: host_name,
            VProps.VITRAGE_TYPE: NOVA_HOST_DATASOURCE,
            VProps.VITRAGE_CATEGORY: EntityCategory.RESOURCE,
            VProps.VITRAGE_SAMPLE_TIMESTAMP: timestamp
        }
        placeholder = \
            host_transformer.create_neighbor_placeholder_vertex(**properties)

        # Test assertions
        observed_uuid = placeholder.vertex_id
        expected_key = tbase.build_key(host_transformer._key_values(
            NOVA_HOST_DATASOURCE,
            host_name))
        expected_uuid = \
            TransformerBase.uuid_from_deprecated_vitrage_id(expected_key)
        self.assertEqual(expected_uuid, observed_uuid)

        observed_time = placeholder.get(VProps.VITRAGE_SAMPLE_TIMESTAMP)
        self.assertEqual(timestamp, observed_time)

        observed_subtype = placeholder.get(VProps.VITRAGE_TYPE)
        self.assertEqual(NOVA_HOST_DATASOURCE, observed_subtype)

        observed_entity_id = placeholder.get(VProps.ID)
        self.assertEqual(host_name, observed_entity_id)

        observed_vitrage_category = placeholder.get(VProps.VITRAGE_CATEGORY)
        self.assertEqual(EntityCategory.RESOURCE, observed_vitrage_category)

        vitrage_is_placeholder = placeholder.get(VProps.VITRAGE_IS_PLACEHOLDER)
        self.assertTrue(vitrage_is_placeholder)

    def test_key_values(self):

        LOG.debug('Test key values')

        # Test setup
        host_name = 'host123456'
        host_transformer = self.transformers[NOVA_HOST_DATASOURCE]

        # Test action
        observed_key_fields = host_transformer._key_values(
            NOVA_HOST_DATASOURCE,
            host_name)

        # Test assertions
        self.assertEqual(EntityCategory.RESOURCE, observed_key_fields[0])
        self.assertEqual(NOVA_HOST_DATASOURCE, observed_key_fields[1])
        self.assertEqual(host_name, observed_key_fields[2])

    def test_snapshot_transform(self):
        LOG.debug('Nova host transformer test: transform entity event')

        # Test setup
        spec_list = mock_sync.simple_host_generators(zone_num=2,
                                                     host_num=4,
                                                     snapshot_events=5)

        host_events = mock_sync.generate_random_events_list(spec_list)

        for event in host_events:
            # Test action
            wrapper = self.transformers[NOVA_HOST_DATASOURCE].transform(event)

            # Test assertions
            self._validate_vertex_props(wrapper.vertex, event)

            neighbors = wrapper.neighbors
            self.assertThat(neighbors, matchers.HasLength(1))
            self._validate_zone_neighbor(neighbors[0], event)

            if DatasourceAction.SNAPSHOT == event[DSProps.DATASOURCE_ACTION]:
                self.assertEqual(GraphAction.UPDATE_ENTITY, wrapper.action)
            else:
                self.assertEqual(GraphAction.CREATE_ENTITY, wrapper.action)

    def _validate_zone_neighbor(self, zone, event):

        zone_name = tbase.extract_field_value(event, 'zone')
        time = event[DSProps.SAMPLE_DATE]

        zt = self.transformers[NOVA_ZONE_DATASOURCE]
        properties = {
            VProps.ID: zone_name,
            VProps.VITRAGE_TYPE: NOVA_ZONE_DATASOURCE,
            VProps.VITRAGE_CATEGORY: EntityCategory.RESOURCE,
            VProps.VITRAGE_SAMPLE_TIMESTAMP: time
        }
        expected_neighbor = \
            zt.create_neighbor_placeholder_vertex(**properties)
        self.assertEqual(expected_neighbor, zone.vertex)

        # Validate neighbor edge
        edge = zone.edge
        transformer = self.transformers[NOVA_HOST_DATASOURCE]
        entity_key = transformer._create_entity_key(event)
        entity_uuid = \
            TransformerBase.uuid_from_deprecated_vitrage_id(entity_key)
        self.assertEqual(edge.source_id, zone.vertex.vertex_id)
        self.assertEqual(edge.target_id, entity_uuid)
        self.assertEqual(edge.label, EdgeLabel.CONTAINS)

    def _validate_vertex_props(self, vertex, event):

        extract_value = tbase.extract_field_value

        expected_id = extract_value(event, 'host')
        observed_id = vertex[VProps.ID]
        self.assertEqual(expected_id, observed_id)
        self.assertEqual(
            EntityCategory.RESOURCE,
            vertex[VProps.VITRAGE_CATEGORY]
        )

        self.assertEqual(
            NOVA_HOST_DATASOURCE,
            vertex[VProps.VITRAGE_TYPE]
        )

        expected_timestamp = event[DSProps.SAMPLE_DATE]
        observed_timestamp = vertex[VProps.VITRAGE_SAMPLE_TIMESTAMP]
        self.assertEqual(expected_timestamp, observed_timestamp)

        expected_name = extract_value(event, 'host')
        observed_name = vertex[VProps.NAME]
        self.assertEqual(expected_name, observed_name)

        vitrage_is_placeholder = vertex[VProps.VITRAGE_IS_PLACEHOLDER]
        self.assertFalse(vitrage_is_placeholder)

        vitrage_is_deleted = vertex[VProps.VITRAGE_IS_DELETED]
        self.assertFalse(vitrage_is_deleted)

    def test_extract_event_action(self):
        LOG.debug('Test extract event action')

        # Test setup
        spec_list = mock_sync.simple_host_generators(
            zone_num=1,
            host_num=1,
            snapshot_events=1,
            snap_vals={DSProps.DATASOURCE_ACTION: DatasourceAction.SNAPSHOT})

        hosts_events = mock_sync.generate_random_events_list(spec_list)
        host_transformer = self.transformers[NOVA_HOST_DATASOURCE]

        # Test action
        action = host_transformer._extract_graph_action(hosts_events[0])

        # Test assertion
        self.assertEqual(GraphAction.UPDATE_ENTITY, action)

        # Test setup
        spec_list = mock_sync.simple_host_generators(
            zone_num=1,
            host_num=1,
            snapshot_events=1,
            snap_vals={DSProps.DATASOURCE_ACTION:
                       DatasourceAction.INIT_SNAPSHOT})
        hosts_events = mock_sync.generate_random_events_list(spec_list)
        host_transformer = self.transformers[NOVA_HOST_DATASOURCE]

        # Test action
        action = host_transformer._extract_graph_action(hosts_events[0])

        # Test assertions
        self.assertEqual(GraphAction.CREATE_ENTITY, action)

        # TODO(lhartal): To add extract action from update event



# Copyright 2018 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log
from oslo_utils import uuidutils

from vitrage.common.constants import TemplateStatus
from vitrage.common.constants import TemplateTypes as TType
from vitrage.common.exception import VitrageError
from vitrage.evaluator.base import Template
from vitrage.evaluator.template_fields import TemplateFields as TFields
from vitrage.evaluator.template_functions.v2 import resolve_parameters
from vitrage.evaluator import template_validation
from vitrage.evaluator.template_validation import base
from vitrage.evaluator.template_validation.content.base import \
    get_content_fault_result
from vitrage.storage.sqlalchemy import models

LOG = log.getLogger(__name__)

METADATA = 'metadata'
NAME = 'name'


def add_templates_to_db(db, templates, template_type, params=None):
    db_rows = list()
    for template in templates:
        final_type = template[METADATA].get(TFields.TYPE, template_type)
        if not final_type or (template_type and template_type != final_type):
            db_rows.append(_get_error_result(template, final_type,
                                             "Unknown template type"))
            continue

        result = _validate_template(db, template, final_type, params)
        if result.is_valid_config:
            result = resolve_parameters(template, params)
            if result.is_valid_config and TFields.PARAMETERS in template:
                # remove the 'parameters' section, it is no longer needed
                del template[TFields.PARAMETERS]

        # validate again, with the resolved parameters
        if result.is_valid_config:
            result = _validate_template(db, template, final_type)

        # template_name might be a parameter, take it after resolve parameters
        template_name = template.get(METADATA).get(NAME)

        if _is_duplicate(db, template_name):
            db_rows.append(_get_error_result(template, final_type,
                                             "Duplicate template name"))
        else:
            db_row = _to_db_row(result, template, final_type)
            db.templates.create(db_row)
            db_rows.append(db_row)

    return db_rows


def validate_templates(db, templates, template_type, params):
    results = list()
    for template in templates:
        final_type = template[METADATA].get(TFields.TYPE, template_type)
        if not final_type or (template_type and template_type != final_type):
            results.append(
                get_content_fault_result(66, "Unknown template type"))
        else:
            results.append(
                _validate_template(db, template, final_type, params))
    return results


def _validate_template(db, template, template_type, params=None):
    if template_type == TType.DEFINITION:
        result = template_validation.validate_definition_template(template)
    elif template_type == TType.STANDARD:
        result = template_validation.validate_template(template,
                                                       _load_def_templates(db),
                                                       params)
    elif template_type == TType.EQUIVALENCE:
        result = base.Result("", True, "", "No Validation")
    else:
        raise VitrageError("Unknown template type %s" % template_type)
    return result


def _is_duplicate(db, template_name):
    if template_name:
        templates = db.templates.query(name=template_name)
        if [t for t in templates if t.status != TemplateStatus.DELETED]:
            return True


def _get_error_result(template, template_type, msg):
    return models.Template(
        name=template[METADATA][NAME],
        uuid="",
        status=TemplateStatus.ERROR,
        status_details=msg,
        file_content=template,
        template_type=template_type)


def _to_db_row(result, template, template_type):
    uuid = uuidutils.generate_uuid()
    status = TemplateStatus.LOADING if result.is_valid_config else \
        TemplateStatus.ERROR
    status_details = result.comment
    return models.Template(
        name=template[METADATA][NAME],
        uuid=uuid,
        status=status,
        status_details=status_details,
        file_content=template,
        template_type=template_type)


def _load_def_templates(db):
    def_templates = {}
    items = db.templates.query(template_type=TType.DEFINITION)
    def_templates_db = [x for x in items if x.status in [
                        TemplateStatus.ACTIVE,
                        TemplateStatus.LOADING]]
    for df in def_templates_db:
        def_templates[df.uuid] = Template(df.uuid,
                                          df.file_content,
                                          df.created_at)
    return def_templates



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import datetime
import json
import zlib

from oslo_db.sqlalchemy import models

from sqlalchemy import Column, DateTime, INTEGER, String, \
    SmallInteger, BigInteger, Index, Boolean, ForeignKey
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship
import sqlalchemy.types as types


DEFAULT_END_TIME = datetime.datetime(2222, 2, 22, 22, 22, 22)


class VitrageBase(models.TimestampMixin, models.ModelBase):
    """Base class for Vitrage Models."""
    __table_args__ = {'mysql_charset': "utf8",
                      'mysql_engine': "InnoDB"}
    __table_initialized__ = False

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __getitem__(self, key):
        return getattr(self, key)

    def update(self, values):
        """Make the model object behave like a dict."""
        for k, v in values.items():
            setattr(self, k, v)


Base = declarative_base(cls=VitrageBase)


class AutoIncrementInteger(types.TypeDecorator):
    impl = types.INT
    count = 0

    def process_bind_param(self, value, dialect):
        value = self.count
        self.count += 1
        return value

    def process_result_value(self, value, dialect):
        return value


MagicBigInt = types.BigInteger().with_variant(AutoIncrementInteger, 'sqlite')


class JSONEncodedDict(types.TypeDecorator):
    """Represents an immutable structure as a json-encoded string"""

    impl = types.TEXT

    def process_bind_param(self, value, dialect):
        return json.dumps(value) if value else None

    def process_result_value(self, value, dialect):
        return json.loads(value) if value else None


class CompressedBinary(types.TypeDecorator):
    impl = types.LargeBinary

    def process_bind_param(self, value, dialect):
        return zlib.compress(value) if value else None

    def process_result_value(self, value, dialect):
        return zlib.decompress(value) if value else None

    def copy(self, **kwargs):
        return CompressedBinary(self.impl.length)


class Event(Base):

    __tablename__ = 'events'

    event_id = Column("id", BigInteger(), primary_key=True, autoincrement=True)
    payload = Column(JSONEncodedDict(), nullable=False)
    is_vertex = Column(Boolean, nullable=False)

    def __repr__(self):
        return \
            "<Event(" \
            "id='%s', " \
            "is_vertex='%s', " \
            "payload='%s')>" % \
            (
                self.event_id,
                self.is_vertex,
                self.payload
            )


class ActiveAction(Base):
    __tablename__ = 'active_actions'
    __table_args__ = (
        # Index 'ix_active_action' on fields:
        # action_type, extra_info, source_vertex_id, target_vertex_id
        Index('ix_active_action', 'action_type', 'extra_info',
              'source_vertex_id', 'target_vertex_id'),
    )

    action_type = Column(String(128))
    extra_info = Column(String(128))
    source_vertex_id = Column(String(128))
    target_vertex_id = Column(String(128))
    action_id = Column(String(128), primary_key=True)
    score = Column(SmallInteger())
    trigger = Column(String(128), primary_key=True)

    def __repr__(self):
        return \
            "<ActiveAction(" \
            "created_at='%s', " \
            "action_type='%s', " \
            "extra_info='%s', " \
            "source_vertex_id='%s', " \
            "target_vertex_id='%s', " \
            "action_id='%s', " \
            "score='%s', " \
            "trigger='%s')>" % \
            (
                self.created_at,
                self.action_type,
                self.extra_info,
                self.source_vertex_id,
                self.target_vertex_id,
                self.action_id,
                self.score,
                self.trigger
            )


class GraphSnapshot(Base):
    __tablename__ = 'graph_snapshots'

    snapshot_id = Column("id", INTEGER, primary_key=True)
    event_id = Column(BigInteger, nullable=False)
    graph_snapshot = Column(CompressedBinary((2 ** 32) - 1), nullable=False)

    def __repr__(self):
        return \
            "<GraphSnapshot(" \
            "id=%s," \
            "event_id='%s', " \
            "graph_snapshot='%s')>" %\
            (
                self.snapshot_id,
                self.event_id,
                self.graph_snapshot
            )


class Template(Base):
    __tablename__ = 'templates'

    uuid = Column("id", String(64), primary_key=True, nullable=False)
    status = Column(String(16))
    status_details = Column(String(128))
    name = Column(String(128), nullable=False)
    file_content = Column(JSONEncodedDict, nullable=False)
    template_type = Column("type", String(64), default='standard')

    def __repr__(self):
        return "<Template(id='%s', name='%s', created_at='%s'," \
               " updated_at='%s', status='%s'," \
               "status_details='%s', file_content='%s', " \
               " template_type='%s' )>" % \
               (self.uuid,
                self.name,
                self.created_at,
                self.updated_at,
                self.status,
                self.status_details,
                self.file_content,
                self.template_type,)


class Webhooks(Base):
    __tablename__ = 'webhooks'

    id = Column(String(128), primary_key=True)
    project_id = Column(String(128), nullable=False)
    is_admin_webhook = Column(Boolean, nullable=False)
    url = Column(String(256), nullable=False)
    headers = Column(String(1024))
    regex_filter = Column(String(512))

    def __repr__(self):
        return \
            "<Webhook(" \
            "id='%s', " \
            "created_at='%s', " \
            "updated_at='%s', " \
            "project_id='%s', " \
            "is_admin_webhook='%s', " \
            "url='%s', " \
            "headers='%s', " \
            "regex_filter='%s')> " %\
            (
                self.id,
                self.created_at,
                self.updated_at,
                self.project_id,
                self.is_admin_webhook,
                self.url,
                self.headers,
                self.regex_filter
            )


class Alarm(Base):

    __tablename__ = 'alarms'

    vitrage_id = Column(String(128), primary_key=True)
    start_timestamp = Column(DateTime, index=True, nullable=False)
    end_timestamp = Column(DateTime, index=True, nullable=False,
                           default=DEFAULT_END_TIME)
    name = Column(String(256), nullable=False)
    vitrage_type = Column(String(64), nullable=False)
    vitrage_aggregated_severity = Column(String(64), index=True,
                                         nullable=False)
    vitrage_operational_severity = Column(String(64), index=True,
                                          nullable=False)
    project_id = Column(String(64), index=True)
    vitrage_resource_type = Column(String(64))
    vitrage_resource_id = Column(String(64))
    vitrage_resource_project_id = Column(String(64), index=True)
    payload = Column(JSONEncodedDict())

    def __repr__(self):
        return \
            "<Alarm(" \
            "vitrage_id='%s', " \
            "start_timestamp='%s', " \
            "end_timestamp='%s'," \
            "name='%s'," \
            "vitrage_type='%s'," \
            "vitrage_aggregated_severity='%s'," \
            "vitrage_operational_severity='%s'," \
            "project_id='%s'," \
            "vitrage_resource_type='%s'," \
            "vitrage_resource_id='%s'," \
            "vitrage_resource_project_id='%s'," \
            "payload='%s')>" % \
            (
                self.vitrage_id,
                self.start_timestamp,
                self.end_timestamp,
                self.name,
                self.vitrage_type,
                self.vitrage_aggregated_severity,
                self.vitrage_operational_severity,
                self.project_id,
                self.vitrage_resource_type,
                self.vitrage_resource_id,
                self.vitrage_resource_project_id,
                self.payload
            )


class Edge(Base):

    __tablename__ = 'edges'

    source_id = Column(String(128),
                       ForeignKey('alarms.vitrage_id', ondelete='CASCADE'),
                       primary_key=True)
    target_id = Column(String(128),
                       ForeignKey('alarms.vitrage_id', ondelete='CASCADE'),
                       primary_key=True)
    label = Column(String(64), nullable=False)
    start_timestamp = Column(DateTime, nullable=False)
    end_timestamp = Column(DateTime, nullable=False, default=DEFAULT_END_TIME)
    payload = Column(JSONEncodedDict())

    source = relationship("Alarm", foreign_keys=[source_id])
    target = relationship("Alarm", foreign_keys=[target_id])

    def __repr__(self):
        return \
            "<Edge(" \
            "source_id='%s', " \
            "target_id='%s', " \
            "label='%s', " \
            "start_timestamp='%s'," \
            "end_timestamp='%s',"\
            "payload='%s)>" % \
            (
                self.source_id,
                self.target_id,
                self.label,
                self.start_timestamp,
                self.end_timestamp,
                self.payload
            )


class Change(Base):

    __tablename__ = 'changes'

    id = Column(MagicBigInt, primary_key=True, autoincrement=True)
    vitrage_id = Column(String(128),
                        ForeignKey('alarms.vitrage_id', ondelete='CASCADE'),
                        index=True, nullable=False)
    timestamp = Column(DateTime, index=True, nullable=False)
    severity = Column(String(64), index=True, nullable=False)
    payload = Column(JSONEncodedDict())

    alarm_id = relationship("Alarm", foreign_keys=[vitrage_id])

    def __repr__(self):
        return \
            "<Change(" \
            "id='%s', " \
            "vitrage_id='%s', " \
            "timestamp='%s', " \
            "severity='%s'," \
            "payload='%s')>" % \
            (
                self.id,
                self.vitrage_id,
                self.timestamp,
                self.severity,
                self.payload
            )



# Copyright 2015 - Alcatel-Lucent
# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import abc
from collections import namedtuple

from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import uuidutils

import vitrage.common.constants as cons
from vitrage.common.constants import DatasourceAction
from vitrage.common.constants import DatasourceOpts as DSOpts
from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.common.constants import EntityCategory
from vitrage.common.constants import GraphAction
from vitrage.common.constants import UpdateMethod
from vitrage.common.constants import VertexProperties as VProps
from vitrage.common.exception import VitrageTransformerError
from vitrage.common.utils import md5
from vitrage.datasources import OPENSTACK_CLUSTER
import vitrage.graph.utils as graph_utils
from vitrage.utils import datetime as datetime_utils
from vitrage.utils import opt_exists

CONF = cfg.CONF
LOG = logging.getLogger(__name__)

EntityWrapper = \
    namedtuple('EntityWrapper', ['vertex', 'neighbors', 'action'])

Neighbor = namedtuple('Neighbor', ['vertex', 'edge'])


TIMESTAMP_FORMAT = '%Y-%m-%dT%H:%M:%SZ'


AVAILABLE = 'available'
CLUSTER_ID = 'OpenStack Cluster'


def extract_field_value(entity_event, *args):
    try:
        value = entity_event
        for key in args:
            value = value[key]

        return value
    except Exception:
        return None


def build_key(key_values):
    return TransformerBase.KEY_SEPARATOR.join(map(str, key_values))


def create_cluster_placeholder_vertex():
    key = build_key([cons.EntityCategory.RESOURCE,
                     OPENSTACK_CLUSTER,
                     CLUSTER_ID])

    metadata = {
        cons.VertexProperties.NAME: OPENSTACK_CLUSTER
    }

    return graph_utils.create_vertex(
        key,
        vitrage_category=cons.EntityCategory.RESOURCE,
        vitrage_type=OPENSTACK_CLUSTER,
        entity_id=CLUSTER_ID,
        entity_state=AVAILABLE,
        metadata=metadata
    )


def convert_timestamp_format(current_timestamp_format, timestamp):

    return datetime_utils.change_time_str_format(
        timestamp,
        current_timestamp_format,
        TIMESTAMP_FORMAT
    )


def is_update_event(event):
    return event[DSProps.DATASOURCE_ACTION] == DatasourceAction.UPDATE


class TransformerBase(object, metaclass=abc.ABCMeta):

    KEY_SEPARATOR = ':'
    QUERY_RESULT = 'graph_query_result'
    METADATA = 'metadata'

    # graph actions which need to refer them differently
    GRAPH_ACTION_MAPPING = {}

    key_to_uuid_cache = {}

    def __init__(self, transformers):
        self.transformers = transformers

    def transform(self, entity_event):
        """Transform an entity event into entity wrapper.

        Entity event is received from driver and it need to be
        transformed into entity wrapper. The wrapper contains:
            1. Entity Vertex - The vertex itself with all fields
            2. Neighbor list - neighbor placeholder vertex and an edge
            3. Action type - CREATE/UPDATE/DELETE

        :param entity_event: a general event from the driver
        :return: entity wrapper
        :rtype:EntityWrapper
        """

        entity_vertex = self._create_entity_vertex(entity_event)
        neighbors = self._create_neighbors(entity_event)
        action = self._extract_graph_action(entity_event)

        if action == GraphAction.DELETE_ENTITY:
            self._delete_id_from_cache(entity_vertex.vertex_id)

        return EntityWrapper(entity_vertex, neighbors, action)

    def _create_entity_vertex(self, entity_event):
        if is_update_event(entity_event) and \
                opt_exists(CONF, self.get_vitrage_type()) and \
                opt_exists(CONF[self.get_vitrage_type()],
                           DSOpts.UPDATE_METHOD):
            update_method = \
                CONF[self.get_vitrage_type()].update_method.lower()
            if update_method == UpdateMethod.PUSH:
                vertex = self._create_update_entity_vertex(entity_event)
                return self.update_uuid_in_vertex(vertex)
            elif update_method == UpdateMethod.PULL:
                vertex = self._create_snapshot_entity_vertex(entity_event)
                return self.update_uuid_in_vertex(vertex)
            elif update_method == UpdateMethod.NONE:
                return None
            else:
                LOG.error('Update event arrived for dataresource that is '
                          'defined without updates')
        else:
            vertex = self._create_snapshot_entity_vertex(entity_event)
            return self.update_uuid_in_vertex(vertex)

    def update_uuid_in_vertex(self, vertex):
        if not vertex:
            return
        # TODO(annarez): remove IS_REAL_VITRAGE_ID prop
        if vertex.get(VProps.IS_REAL_VITRAGE_ID):
            return vertex
        new_uuid = self.uuid_from_deprecated_vitrage_id(vertex.vertex_id)
        vertex.properties[VProps.VITRAGE_CACHED_ID] = md5(vertex.vertex_id)
        vertex.vertex_id = new_uuid
        vertex.properties[VProps.VITRAGE_ID] = new_uuid
        vertex.properties[VProps.IS_REAL_VITRAGE_ID] = True
        return vertex

    @classmethod
    def uuid_from_deprecated_vitrage_id(cls, vitrage_id):
        old_vitrage_id = md5(vitrage_id)
        new_uuid = cls.key_to_uuid_cache.get(old_vitrage_id)
        if not new_uuid:
            new_uuid = uuidutils.generate_uuid()
            cls.key_to_uuid_cache[old_vitrage_id] = new_uuid

        return new_uuid

    @classmethod
    def _delete_id_from_cache(cls, vitrage_id):
        for key, value in cls.key_to_uuid_cache.items():
            if value == vitrage_id:
                del cls.key_to_uuid_cache[key]
                break

    @abc.abstractmethod
    def _create_snapshot_entity_vertex(self, entity_event):
        pass

    @abc.abstractmethod
    def _create_update_entity_vertex(self, entity_event):
        pass

    def _create_neighbors(self, entity_event):
        """Extracts entity neighbors received from a given entity event.

         Extracting entity neighbors from a given event provided
         by driver

         :param entity_event: an event provided by driver
         :return: neighbors - a list of neighbors where each item in the list
                              is a tuple of (vertex, edge)
         :rtype: list
         """

        if is_update_event(entity_event):
            return self._create_update_neighbors(entity_event)
        else:
            return self._create_snapshot_neighbors(entity_event)

    def _create_snapshot_neighbors(self, entity_event):
        return []

    def _create_update_neighbors(self, entity_event):
        return []

    @abc.abstractmethod
    def _create_entity_key(self, entity_event):
        """Create an entity key from given event

        By given an entity event, return an entity key which
        consists of key fields

        :param entity_event: event that returns from the driver
        :return: key
        """
        pass

    def _create_neighbor(self,
                         entity_event,
                         neighbor_id,
                         neighbor_datasource_type,
                         relationship_type,
                         neighbor_category=EntityCategory.RESOURCE,
                         is_entity_source=True,
                         metadata=None):
        metadata = {} if metadata is None else metadata
        # create placeholder vertex
        entity_vitrage_id = \
            self.uuid_from_deprecated_vitrage_id(
                self._create_entity_key(entity_event))
        vitrage_sample_timestamp = entity_event[DSProps.SAMPLE_DATE]
        properties = {
            VProps.ID: neighbor_id,
            VProps.VITRAGE_TYPE: neighbor_datasource_type,
            VProps.VITRAGE_CATEGORY: neighbor_category,
            VProps.VITRAGE_SAMPLE_TIMESTAMP: vitrage_sample_timestamp,
            self.METADATA: metadata
        }
        neighbor_vertex = \
            self.create_neighbor_placeholder_vertex(**properties)
        # connect placeholder vertex to entity vertex
        edge_direction = self._get_edge_direction(entity_vitrage_id,
                                                  neighbor_vertex.vertex_id,
                                                  is_entity_source)
        relationship_edge = graph_utils.create_edge(
            source_id=edge_direction[0],
            target_id=edge_direction[1],
            relationship_type=relationship_type)

        return Neighbor(neighbor_vertex, relationship_edge)

    @staticmethod
    def _get_edge_direction(entity_id,
                            neighbor_id,
                            is_entity_source):
        source_id = entity_id
        target_id = neighbor_id

        if not is_entity_source:
            source_id = neighbor_id
            target_id = entity_id

        return source_id, target_id

    def _key_values(self, *args):
        return (EntityCategory.RESOURCE,) + args

    def create_neighbor_placeholder_vertex(self, **kwargs):
        if VProps.VITRAGE_TYPE not in kwargs:
            LOG.error("Can't create placeholder vertex. Missing property TYPE")
            raise ValueError('Missing property TYPE')

        if VProps.ID not in kwargs:
            LOG.error("Can't create placeholder vertex. Missing property ID")
            raise ValueError('Missing property ID')

        metadata = {}
        if self.METADATA in kwargs:
            metadata = kwargs[self.METADATA]

        key_fields = self._key_values(kwargs[VProps.VITRAGE_TYPE],
                                      kwargs[VProps.ID])

        vertex = graph_utils.create_vertex(
            build_key(key_fields),
            vitrage_category=kwargs[VProps.VITRAGE_CATEGORY],
            vitrage_type=kwargs[VProps.VITRAGE_TYPE],
            vitrage_sample_timestamp=kwargs[VProps.VITRAGE_SAMPLE_TIMESTAMP],
            vitrage_is_placeholder=kwargs.get(VProps.VITRAGE_IS_PLACEHOLDER,
                                              True),
            entity_id=kwargs[VProps.ID],
            metadata=metadata)
        return self.update_uuid_in_vertex(vertex)

    def _extract_graph_action(self, entity_event):
        """Extract graph action.

        Decides what action (from constants.GraphAction) the processor need
        to perform according to the data received from the event.

        :param entity_event: event that returns from the driver
        :return: the action that the processor needs to perform
        :rtype: str
        """
        if DSProps.EVENT_TYPE in entity_event and \
                entity_event[DSProps.EVENT_TYPE] in vars(GraphAction).values():
            return entity_event[DSProps.EVENT_TYPE]

        datasource_action = entity_event[DSProps.DATASOURCE_ACTION]

        if DatasourceAction.UPDATE == datasource_action:
            return self.GRAPH_ACTION_MAPPING.get(
                entity_event.get(DSProps.EVENT_TYPE, None),
                GraphAction.UPDATE_ENTITY)

        if DatasourceAction.SNAPSHOT == datasource_action:
            return GraphAction.UPDATE_ENTITY

        if DatasourceAction.INIT_SNAPSHOT == datasource_action:
            return GraphAction.CREATE_ENTITY

        raise VitrageTransformerError(
            'Invalid action type: (%s)' % datasource_action)

    @staticmethod
    def _format_update_timestamp(update_timestamp, sample_timestamp):
        update_timestamp = update_timestamp if update_timestamp \
            else sample_timestamp
        return datetime_utils.format_timestamp(update_timestamp)

    @staticmethod
    def get_enrich_query(event):
        """Allow running a query on the graph vertices

        The result of the query specified here will be added to the event in
        the 'QUERY_RESULT' field.

         Example:
         -------
          query = {'vitrage_type': 'nova.instance'}
          Before transform is called the result of running the query against
          the topology graph will be updated to event[QUERY_RESULT]
          To contain the list of all the vertices with
          vitrage_type=nova.instance
        """
        return None

    @abc.abstractmethod
    def get_vitrage_type(self):
        """Returns the vitrage_type of the datasource

        :return: datasource type
        :rtype: String
        """
        pass



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import math
import os
from os import path
import sys

import pecan

from distutils import spawn
from oslo_config import cfg
from oslo_log import log
from oslo_utils import uuidutils
# noinspection PyPackageRequirements
from paste import deploy
from werkzeug import serving

from vitrage.api import hooks

CONF = cfg.CONF

LOG = log.getLogger(__name__)

# NOTE(sileht): pastedeploy uses ConfigParser to handle
# global_conf, since python 3 ConfigParser doesn't
# allow storing object as config value, only strings are
# permit, so to be able to pass an object created before paste load
# the app, we store them into a global var. But the each loaded app
# store it's configuration in unique key to be concurrency safe.
global APPCONFIGS
APPCONFIGS = {}


def setup_app(root, conf=None):
    app_hooks = [hooks.ConfigHook(),
                 hooks.TranslationHook(),
                 hooks.GCHook(),
                 hooks.RPCHook(),
                 hooks.ContextHook(),
                 hooks.DBHook(),
                 hooks.CoordinatorHook()]

    app = pecan.make_app(
        root,
        hooks=app_hooks,
        guess_content_type_from_ext=False
    )

    return app


def load_app():
    global APPCONFIGS
    # Build the WSGI app
    cfg_path = CONF.api.paste_config
    if not os.path.isabs(cfg_path):
        cfg_path = CONF.find_file(cfg_path)

    if cfg_path is None or not os.path.exists(cfg_path):
        raise cfg.ConfigFilesNotFoundError([CONF.api.paste_config])

    config = dict(conf=CONF)
    configkey = uuidutils.generate_uuid()
    APPCONFIGS[configkey] = config

    LOG.info('Full WSGI config used: %s', cfg_path)

    appname = "vitrage+" + CONF.api.auth_mode
    return deploy.loadapp("config:" + cfg_path, name=appname,
                          global_conf={'configkey': configkey})


def build_server():
    uwsgi = spawn.find_executable("uwsgi")
    if not uwsgi:
        LOG.warning('uwsgi not installed, starting a TEST server')
        build_simple_server()
    else:
        build_uwsgi_server(uwsgi)


def wsgi_file():
    return path.join(path.dirname(__file__), 'app.wsgi')


def build_uwsgi_server(uwsgi):

    major = str(sys.version_info.major)
    minor = str(sys.version_info.minor)
    plugin = ''.join(('python', major, minor))
    args = [
        "--if-not-plugin", "python", "--plugin", plugin, "--endif",
        "--http-socket", "%s:%d" % (CONF.api.host, CONF.api.port),
        "--master",
        "--enable-threads",
        "--thunder-lock",
        "--hook-master-start", "unix_signal:15 gracefully_kill_them_all",
        "--die-on-term",
        "--processes", str(math.floor(CONF.api.workers * 1.5)),
        "--threads", str(CONF.api.workers),
        "--lazy-apps",
        "--chdir", "/",
        "--buffer-size", "65535",
        "--wsgi-file", wsgi_file(),
        "--procname-prefix", "vitrage",
        "--pyargv", " ".join(sys.argv[1:]),
    ]

    virtual_env = os.getenv("VIRTUAL_ENV")
    if virtual_env is not None:
        args.extend(["-H", os.getenv("VIRTUAL_ENV", ".")])

    return os.execl(uwsgi, uwsgi, *args)


def build_simple_server():
    app = load_app()
    # Create the WSGI server and start it
    host, port = CONF.api.host, CONF.api.port

    LOG.info('Starting server in PID %s', os.getpid())
    LOG.info('Configuration:')
    CONF.log_opt_values(LOG, log.INFO)

    if host == '0.0.0.0':
        LOG.info(
            'serving on 0.0.0.0:%(port)s, view at http://127.0.0.1:%(port)s',
            {'port': port})
    else:
        LOG.info('serving on http://%(host)s:%(port)s',
                 {'host': host, 'port': port})

    LOG.info('"DANGER! For testing only, do not use in production"')

    serving.run_simple(host, port,
                       app, processes=CONF.api.workers)


def app_factory(global_config, **local_conf):
    global APPCONFIGS
    appconfig = APPCONFIGS.get(global_config.get('configkey'))
    return setup_app(root=local_conf.get('root'), **appconfig)



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
import functools

import itertools

from vitrage.graph.driver.elements import Vertex


def _before_func(graph, item):
    if not graph.is_subscribed():
        return
    return graph.get_item(item)


def _after_func(graph, item, data_before=None):
    if not graph.is_subscribed():
        return
    element = graph.get_item(item)
    is_vertex = isinstance(element, Vertex) or isinstance(item, Vertex)
    graph.notifier.notify(data_before, element, is_vertex, graph)


class Notifier(object):
    def __init__(self):
        self._subscriptions = []
        self._finalization_subscriptions = []

    def subscribe(self, function, finalization=False):
        if finalization:
            self._finalization_subscriptions.append(function)
        else:
            self._subscriptions.append(function)

    def is_subscribed(self):
        size = len(self._subscriptions) + len(self._finalization_subscriptions)
        return size != 0

    def notify(self, *args, **kwargs):
        for func in itertools.chain(self._subscriptions,
                                    self._finalization_subscriptions):
            func(*args, **kwargs)

    @staticmethod
    def update_notify(func):
        @functools.wraps(func)
        def notified_func(graph, item, *args, **kwargs):
            data_before = _before_func(graph, item)
            func(graph, item, *args, **kwargs)
            _after_func(graph, item, data_before)
        return notified_func

    @staticmethod
    def add_notify(func):
        @functools.wraps(func)
        def notified_func(graph, item, *args, **kwargs):
            func(graph, item, *args, **kwargs)
            _after_func(graph, item)
        return notified_func



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from dateutil import parser
import json

from oslo_log import log
from osprofiler import profiler

from vitrage.api_handler.apis import base
from vitrage.common.constants import EntityCategory as ECategory
from vitrage.common.constants import HistoryProps as HProps
from vitrage.common.constants import TenantProps
from vitrage.common.constants import VertexProperties as VProps
from vitrage.common.utils import compress_obj
from vitrage.datasources.alarm_properties import AlarmProperties as AProps
from vitrage.entity_graph.mappings.operational_alarm_severity import \
    OperationalAlarmSeverity
from vitrage.storage import db_time

LOG = log.getLogger(__name__)


@profiler.trace_cls("alarm apis",
                    info={}, hide_args=False, trace_private=False)
class AlarmApis(base.EntityGraphApisBase):

    @base.lock_graph
    def get_alarms(self, ctx, vitrage_id, all_tenants, *args, **kwargs):

        kwargs = self._parse_kwargs(kwargs)

        if not vitrage_id or vitrage_id == 'all':
            if not all_tenants:
                kwargs['project_id'] = \
                    ctx.get(TenantProps.PROJECT_ID, 'no-project')
                kwargs['is_admin_project'] = \
                    ctx.get(TenantProps.IS_ADMIN, False)
        else:
            kwargs.get('filter_by', []).append(VProps.VITRAGE_RESOURCE_ID)
            kwargs.get('filter_vals', []).append(vitrage_id)

        alarms = self._get_alarms(*args, **kwargs)
        data = {'alarms': [v.payload for v in alarms]}
        return compress_obj(data, level=1)

    # TODO(annarez): add db support
    @base.lock_graph
    def show_alarm(self, ctx, vitrage_id):
        LOG.debug('Show alarm with vitrage_id: %s', vitrage_id)

        alarm = self.entity_graph.get_vertex(vitrage_id)
        if not alarm or alarm.get(VProps.VITRAGE_CATEGORY) != ECategory.ALARM:
            LOG.warning('Alarm show - Not found (%s)', vitrage_id)
            return None

        is_admin = ctx.get(TenantProps.IS_ADMIN, False)
        curr_project = ctx.get(TenantProps.PROJECT_ID, None)
        alarm_project = alarm.get(VProps.PROJECT_ID)
        if not is_admin and curr_project != alarm_project:
            LOG.warning('Alarm show - Authorization failed (%s)', vitrage_id)
            return None

        return json.dumps(alarm.properties)

    @base.lock_graph
    def get_alarm_counts(self, ctx, all_tenants):
        LOG.debug("AlarmApis get_alarm_counts - all_tenants=%s", all_tenants)

        project_id = ctx.get(TenantProps.PROJECT_ID, None)
        is_admin_project = ctx.get(TenantProps.IS_ADMIN, False)

        if all_tenants:
            counts = self.db.history_facade.count_active_alarms()

        else:
            counts = self.db.history_facade.count_active_alarms(
                project_id=project_id,
                is_admin_project=is_admin_project)

        return json.dumps(counts)

    def _get_alarms(self, *args, **kwargs):
        """Finds all the alarms with project_id

        Finds all the alarms which has the project_id. In case the tenant is
        admin then project_id can also be None.

        :rtype: list
        """
        alarms = self.db.history_facade.get_alarms(*args, **kwargs)

        for alarm in alarms:
            start_timestamp = \
                self.db.history_facade.add_utc_timezone(alarm.start_timestamp)
            alarm.payload[HProps.START_TIMESTAMP] = str(start_timestamp)
            if alarm.end_timestamp <= db_time():
                end_timestamp = \
                    self.db.history_facade.add_utc_timezone(
                        alarm.end_timestamp)
                alarm.payload[HProps.END_TIMESTAMP] = str(end_timestamp)
                # change operational severity of ended alarms to 'OK'
                # TODO(annarez): in next version use only 'state'
                alarm.payload[VProps.VITRAGE_OPERATIONAL_SEVERITY] = \
                    OperationalAlarmSeverity.OK
                # TODO(annarez): implement state change in processor and DB
                alarm.payload[VProps.STATE] = AProps.INACTIVE_STATE

        return alarms

    def _parse_kwargs(self, kwargs):
        kwargs = {k: v for k, v in kwargs.items() if v is not None}

        if kwargs.get('start'):
            kwargs['start'] = parser.parse(kwargs['start'])
        if kwargs.get('end'):
            kwargs['end'] = parser.parse(kwargs['end'])
        if kwargs.get('sort_by') and type(kwargs.get('sort_by')) != list:
            kwargs['sort_by'] = [kwargs.get('sort_by')]
        if kwargs.get('sort_dirs') and type(kwargs.get('sort_dirs')) != list:
            kwargs['sort_dirs'] = [kwargs.get('sort_dirs')]
        if str(kwargs.get('next_page')).lower() == 'false':
            kwargs['next_page'] = False
        else:
            kwargs['next_page'] = True

        if kwargs.get('filter_by') and type(kwargs.get('filter_by')) != list:
            kwargs['filter_by'] = [kwargs.get('filter_by')]
        if kwargs.get('filter_vals') and type(
                kwargs.get('filter_vals')) != list:
            kwargs['filter_vals'] = [kwargs.get('filter_vals')]

        return kwargs



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
from oslo_log import log as logging

from vitrage.common.constants import NotifierEventTypes
from vitrage.common.constants import VertexProperties as VProps
from vitrage.datasources.aodh import AODH_DATASOURCE
from vitrage.datasources.aodh.properties import AodhState
from vitrage.entity_graph.mappings.operational_alarm_severity import \
    OperationalAlarmSeverity
from vitrage.notifier.plugins.base import NotifierBase
from vitrage import os_clients


LOG = logging.getLogger(__name__)


severity_translation = {
    OperationalAlarmSeverity.CRITICAL: 'critical',
    OperationalAlarmSeverity.SEVERE: 'moderate',
    OperationalAlarmSeverity.WARNING: 'low',
}


class AodhNotifier(NotifierBase):

    @staticmethod
    def get_notifier_name():
        return AODH_DATASOURCE

    def __init__(self):
        super(AodhNotifier, self).__init__()
        self.client = os_clients.aodh_client()

    def process_event(self, data, event_type):
        response = None
        if event_type == NotifierEventTypes.ACTIVATE_DEDUCED_ALARM_EVENT:
            if not data.get(VProps.ID):
                response = self._create_aodh_alarm(data, AodhState.ALARM)
            else:
                response = self._update_aodh_alarm(data, AodhState.ALARM)
        elif event_type == NotifierEventTypes.DEACTIVATE_DEDUCED_ALARM_EVENT:
            response = self._update_aodh_alarm(data, AodhState.OK)

        if response and response.alarm_id:
            LOG.info('Aodh Alarm id %s: ', response.alarm_id)
        else:
            LOG.error('Failed to %s Aodh Alarm \n%s', event_type, data)

    def _create_aodh_alarm(self, alarm, state):
        alarm_request = _alarm_request(alarm, state)
        try:
            LOG.info('Aodh Alarm - Activate: ' + alarm_request)
            return self.client.alarm.create(alarm_request)
        except Exception:
            LOG.exception('Failed to activate Aodh Alarm. Got Exception.')

    def _update_aodh_alarm(self, alarm, state):
        aodh_id = alarm.get(VProps.ID)
        try:
            LOG.info('Aodh Alarm $%s update state %s', aodh_id, state)
            alarm_update = {AodhState: state}
            return self.client.alarm.update(alarm_id=aodh_id,
                                            alarm_update=alarm_update)
        except Exception:
            LOG.exception('Failed to update Aodh Alarm. Got Exception.')


def _alarm_request(data, state):
    affected_resource_id = data.get(VProps.RESOURCE, {}).get(VProps.ID, '')
    alarm_name = data.get(VProps.NAME)
    aodh_alarm_name = '_'.join([alarm_name, affected_resource_id])
    severity = severity_translation.get(data.get(VProps.SEVERITY), 'low')
    return dict(
        name=aodh_alarm_name,
        description='Vitrage deduced alarm',
        event_rule=dict(query=[
            dict(
                field='resource_id',
                type='',
                op='eq',
                value=affected_resource_id),
            dict(
                field='vitrage_id',
                type='',
                op='eq',
                value=data.get(VProps.VITRAGE_ID))]),
        severity=severity,
        state=state,
        type='event')



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from vitrage.evaluator.actions.base import ActionType
from vitrage.evaluator.template_fields import TemplateFields
from vitrage.evaluator.template_validation.content.v1.\
    add_causal_relationship_validator import AddCausalRelationshipValidator
from vitrage.tests.unit.evaluator.template_validation.content.base import \
    ActionValidatorTest
from vitrage.tests.unit.evaluator.template_validation.content.base import \
    DEFINITIONS_INDEX_MOCK


class AddCausalRelationshipValidatorTest(ActionValidatorTest):

    def test_validate_add_causal_relationship_action(self):

        self._validate_action(
            self._create_add_causal_relationship_action('a1', 'a2'),
            AddCausalRelationshipValidator.validate
        )

    def test_validate_add_causal_relation_action_without_action_target(self):

        self._validate_action_without_action_target(
            self._create_add_causal_relationship_action('a1', 'a2'),
            AddCausalRelationshipValidator.validate
        )

    def test_validate_add_causal_relationship_action_with_invalid_target(self):

        self._validate_action_with_invalid_target_id(
            self._create_add_causal_relationship_action('unknown', 'a2'),
            AddCausalRelationshipValidator.validate
        )

    def test_validate_add_causal_relationship_action_with_invalid_source(self):

        self._validate_action_with_invalid_target_id(
            self._create_add_causal_relationship_action('a1', 'unknown'),
            AddCausalRelationshipValidator.validate
        )

    def test_validate_add_causal_relationship_action_without_target(self):

        self._validate_action_without_target_id(
            self._create_add_causal_relationship_action('a1', 'a2'),
            AddCausalRelationshipValidator.validate,
            130
        )

    def test_validate_add_causal_relationship_action_without_source(self):

        # Test setup
        idx = DEFINITIONS_INDEX_MOCK.copy()
        action = self._create_add_causal_relationship_action('a1', 'a2')
        action[TemplateFields.ACTION_TARGET].pop(TemplateFields.SOURCE, None)

        # Test action
        result = AddCausalRelationshipValidator.validate(action, idx)

        # Test assertion
        self._assert_fault_result(result, 130)

    def test_validate_add_causal_relationship_action_wrong_src_category(self):

        # Test setup
        idx = DEFINITIONS_INDEX_MOCK.copy()
        action = self._create_add_causal_relationship_action('a1', '123')

        # Test action
        result = AddCausalRelationshipValidator.validate(action, idx)

        # Test assertion
        self._assert_fault_result(result, 132)

    def test_validate_add_causal_relationship_action_wrong_tgt_category(self):

        # Test setup
        idx = DEFINITIONS_INDEX_MOCK.copy()
        action = self._create_add_causal_relationship_action('123', 'a1')

        # Test action
        result = AddCausalRelationshipValidator.validate(action, idx)

        # Test assertion
        self._assert_fault_result(result, 132)

    @staticmethod
    def _create_add_causal_relationship_action(target, source):

        action_target = {
            TemplateFields.TARGET: target,
            TemplateFields.SOURCE: source
        }
        action = {
            TemplateFields.ACTION_TYPE: ActionType.ADD_CAUSAL_RELATIONSHIP,
            TemplateFields.ACTION_TARGET: action_target}

        return action



# Copyright 2020
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_config import cfg
from oslo_log import log as logging

from testtools import matchers

from vitrage.common.constants import DatasourceAction
from vitrage.common.constants import DatasourceOpts as DSOpts
from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.common.constants import UpdateMethod

from vitrage.datasources.tmfapi639 import TMFAPI639_DATASOURCE
from vitrage.datasources.tmfapi639.transformer import TmfApi639Transformer
from vitrage.datasources import transformer_base
from vitrage.datasources.transformer_base import TransformerBase

from vitrage.tests.unit.datasources.test_transformer_base import \
    BaseTransformerTest

from datetime import datetime
from json import loads

LOG = logging.getLogger(__name__)

message = '[{"id":"1","name":"Host-1","@type":"Host",\
    "resourceRelationship":[{"type":"parent","resource":{"id":"1"}}]},\
            {"id":"2","name":"Host-2","@type":"Host",\
            "resourceRelationship":[{"type":"parent","resource":{"id":"1"}}]}]'


# noinspection PyProtectedMember
class TestTmfApi639Transformer(BaseTransformerTest):

    OPTS = [
        cfg.StrOpt(DSOpts.UPDATE_METHOD,
                   default=UpdateMethod.PULL),
    ]

    # noinspection PyAttributeOutsideInit,PyPep8Naming
    @classmethod
    def setUpClass(cls):
        super(TestTmfApi639Transformer, cls).setUpClass()
        cls.transformers = {}
        cls.conf = cfg.ConfigOpts()
        cls.conf.register_opts(cls.OPTS, group=TMFAPI639_DATASOURCE)
        cls.transformer = TmfApi639Transformer(cls.transformers)
        cls.transformers[TMFAPI639_DATASOURCE] = cls.transformer

    # noinspection PyAttributeOutsideInit
    def setUp(self):
        super(TestTmfApi639Transformer, self).setUp()
        self.timestamp = datetime.utcnow()

    def test_create_entity_key(self):
        event = loads(message)[0]
        self.assertIsNotNone(event)

        transformer = TmfApi639Transformer(self.transformers)
        observed_key = transformer._create_entity_key(event)

        entity_type = TMFAPI639_DATASOURCE
        entity_id = event["id"]

        # Test assertions
        observed_key_fields = observed_key.split(
            TransformerBase.KEY_SEPARATOR)

        self.assertEqual(entity_type, observed_key_fields[1])
        self.assertEqual(entity_id, observed_key_fields[2])

    # Transformer tests:
    # - Vertex creation
    # - Neighbor link

    def test_topology(self):

        sample_timestamp = \
            datetime.now().strftime(transformer_base.TIMESTAMP_FORMAT)
        update_timestamp = TransformerBase._format_update_timestamp(
            update_timestamp=None,
            sample_timestamp=sample_timestamp)

        transformer = self.transformers[TMFAPI639_DATASOURCE]

        # Create 1 vertex
        event1 = loads(message)[0]
        event1[DSProps.DATASOURCE_ACTION] = DatasourceAction.SNAPSHOT
        event1[DSProps.SAMPLE_DATE] = update_timestamp
        self.assertIsNotNone(event1)

        # Create vertex 1
        wrapper1 = transformer.transform(event1)
        # Assertion
        self._validate_base_vertex_props(
            wrapper1.vertex,
            event1["name"],
            TMFAPI639_DATASOURCE
        )

        # Create 2nd vertex
        event2 = loads(message)[1]
        event2[DSProps.DATASOURCE_ACTION] = DatasourceAction.SNAPSHOT
        event2[DSProps.SAMPLE_DATE] = update_timestamp
        self.assertIsNotNone(event2)

        # Create vertex 2
        wrapper2 = transformer.transform(event2)
        # Assertion
        self._validate_base_vertex_props(
            wrapper2.vertex,
            event2["name"],
            TMFAPI639_DATASOURCE
        )

        # Test whether they are linked
        self.assertThat(wrapper2.neighbors, matchers.HasLength(1))

        parent_id = transformer._create_entity_key(event1)
        parent_uuid = \
            transformer.uuid_from_deprecated_vitrage_id(parent_id)

        child_id = transformer._create_entity_key(event2)
        child_uuid = \
            transformer.uuid_from_deprecated_vitrage_id(child_id)

        self.assertEqual(wrapper2.neighbors[0].edge.source_id, child_uuid)
        self.assertEqual(wrapper2.neighbors[0].edge.target_id, parent_uuid)



# Copyright 2015 - Alcatel-Lucent
# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import itertools
import os

from oslo_config import cfg
from oslo_log import log
from oslo_utils import importutils

import vitrage.api
import vitrage.coordination
import vitrage.datasources
import vitrage.entity_graph.consistency
import vitrage.evaluator
import vitrage.evaluator.actions
import vitrage.keystone_client
import vitrage.machine_learning
import vitrage.machine_learning.plugins.jaccard_correlation
import vitrage.notifier
import vitrage.notifier.plugins.snmp
import vitrage.notifier.plugins.webhook
import vitrage.notifier.plugins.zaqar
import vitrage.os_clients
import vitrage.persistency
import vitrage.rpc
import vitrage.snmp_parsing
import vitrage.storage

CONF = cfg.CONF
LOG = log.getLogger(__name__)

DATASOURCES_PATH = 'vitrage.datasources.'
DATASOURCE_FS_PATH = os.path.join('vitrage', 'datasources')
DRIVER_FILE = 'driver.py'
TRANSFORMER_FILE = 'alarm_transformer_base.py'


def list_opts():
    return [
        ('api', vitrage.api.OPTS),
        ('datasources', vitrage.datasources.OPTS),
        ('evaluator', vitrage.evaluator.OPTS),
        ('evaluator_actions', vitrage.evaluator.actions.OPTS),
        ('consistency', vitrage.entity_graph.consistency.OPTS),
        ('database', vitrage.storage.OPTS),
        ('persistency', vitrage.persistency.OPTS),
        ('entity_graph', vitrage.entity_graph.OPTS),
        ('service_credentials', vitrage.keystone_client.OPTS),
        ('machine_learning',
         vitrage.machine_learning.OPTS),
        ('jaccard_correlation',
         vitrage.machine_learning.plugins.jaccard_correlation.OPTS),
        ('snmp', vitrage.notifier.plugins.snmp.OPTS),
        ('webhook', vitrage.notifier.plugins.webhook.OPTS),
        ('snmp_parsing', vitrage.snmp_parsing.OPTS),
        ('zaqar', vitrage.notifier.plugins.zaqar.OPTS),
        ('coordination', vitrage.coordination.OPTS),
        ('DEFAULT', itertools.chain(
            vitrage.os_clients.OPTS,
            vitrage.rpc.OPTS,
            vitrage.notifier.OPTS))
    ]


def datasources_opts():

    top = os.getcwd()

    datasources = _normalize_path_to_datasource_name(
        _filter_folders_containing_transformer(_get_datasources_folders(top)),
        top)

    return [(datasource, module.OPTS) for datasource in datasources
            for module in
            [importutils.import_module(DATASOURCES_PATH + datasource)]
            if 'OPTS' in vars(module)]


def _get_datasources_folders(top=os.getcwd()):
    return [os.path.dirname(os.path.join(root, name))
            for root, dirs, files in os.walk(top, topdown=False)
            for name in files if name == DRIVER_FILE]


def _filter_folders_containing_transformer(folders):
    return [folder for folder in folders for
            root, dirs, files in os.walk(folder, topdown=False) for
            name in files if name == TRANSFORMER_FILE]


def _normalize_path_to_datasource_name(path_list, top=os.getcwd()):
    return [os.path.relpath(path, os.path.join(top, DATASOURCE_FS_PATH))
            .replace(os.sep, '.') for path in path_list]


def register_opts(package_name, paths):
    """register opts of package package_name, with base path in paths"""
    for path in paths:
        try:
            opt = importutils.import_module(
                "%s.%s" % (path, package_name)).OPTS
            CONF.register_opts(
                list(opt),
                group=None if package_name == 'DEFAULT' else package_name
            )
            return
        except ImportError:
            pass
    LOG.error("Failed to import config options for %s. Not found in %s",
              package_name, paths)



# Copyright 2018 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR  CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from kubernetes import client
from kubernetes import config
from oslo_config import cfg
from oslo_log import log
from vitrage.datasources.driver_base import DriverBase
from vitrage.datasources.kubernetes.properties import KUBERNETES_DATASOURCE
from vitrage.datasources.kubernetes.properties import KubernetesProperties\
    as kubProp

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class KubernetesDriver(DriverBase):

    def __init__(self):
        super(KubernetesDriver, self).__init__()
        self._client = None

    @property
    def client(self):
        if not self._client:
            self._client = self._k8s_client()
        return self._client

    @staticmethod
    def _k8s_client():
        try:
            if not CONF.kubernetes.config_file:
                LOG.warning('kubernetes config file is not defined')
                return

            kubeconf = CONF.kubernetes.config_file
            config.load_kube_config(config_file=kubeconf)
            k8s_client = client.CoreV1Api()
            if k8s_client is None:
                LOG.warning('k8s client returns None')
                return

            return k8s_client
        except Exception:
            LOG.exception('Create k8s client - Got Exception')

    def get_all(self, datasource_action):
        return self.make_pickleable(self._prepare_entities(
                                    self.client.list_node()),
                                    KUBERNETES_DATASOURCE,
                                    datasource_action)

    @staticmethod
    def _prepare_entities(nodes):
        entities = []
        for item in nodes.items:
            metadata = item.metadata
            node_details = {
                kubProp.NAME: metadata.name,
                kubProp.UID: metadata.uid,
                kubProp.CREATION_TIMESTAMP: metadata.creation_timestamp,
                kubProp.EXTERNALID: item.spec.external_id,
                kubProp.PROVIDER_NAME: item.spec.provider_id.split(":///")[0],
                }
            entities.append(node_details)
        return [{'resources': entities}]



# Copyright 2016 - Alcatel-Lucent
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,  software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND,  either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from vitrage.datasources.nova.nova_driver_base import NovaDriverBase
from vitrage.datasources.nova.zone import NOVA_ZONE_DATASOURCE


class ZoneDriver(NovaDriverBase):

    @staticmethod
    def filter_internal_zone(zones):
        zones_res = []
        for zone in zones:
            zone_dict = vars(zone)
            if zone_dict['zoneName'] and zone_dict['zoneName'] != 'internal':
                zones_res.append(zone_dict)
        return zones_res

    def get_all(self, datasource_action):
        return self.make_pickleable(self.filter_internal_zone(
            self.client.availability_zones.list()),
            NOVA_ZONE_DATASOURCE,
            datasource_action,
            *self.properties_to_filter_out())

    @staticmethod
    def properties_to_filter_out():
        return ['manager', '_info']



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.common.constants import EdgeLabel
from vitrage.common.constants import EntityCategory
from vitrage.common.constants import GraphAction
from vitrage.common.constants import VertexProperties as VProps
from vitrage.datasources.cinder.volume import CINDER_VOLUME_DATASOURCE
from vitrage.datasources.heat.stack import HEAT_STACK_DATASOURCE
from vitrage.datasources.heat.stack.properties import StackProperties\
    as StackProps
from vitrage.datasources.neutron.network import NEUTRON_NETWORK_DATASOURCE
from vitrage.datasources.neutron.port import NEUTRON_PORT_DATASOURCE
from vitrage.datasources.nova.instance import NOVA_INSTANCE_DATASOURCE
from vitrage.datasources.resource_transformer_base import \
    ResourceTransformerBase
from vitrage.datasources import transformer_base as tbase
from vitrage.datasources.transformer_base import build_key
from vitrage.datasources.transformer_base import extract_field_value
import vitrage.graph.utils as graph_utils


class HeatStackTransformer(ResourceTransformerBase):

    RESOURCE_TYPE = {
        'OS::Nova::Server': NOVA_INSTANCE_DATASOURCE,
        'OS::Cinder::Volume': CINDER_VOLUME_DATASOURCE,
        'OS::Neutron::Net': NEUTRON_NETWORK_DATASOURCE,
        'OS::Neutron::Port': NEUTRON_PORT_DATASOURCE
    }

    # graph actions which need to refer them differently
    GRAPH_ACTION_MAPPING = {
        'orchestration.stack.delete.end': GraphAction.DELETE_ENTITY,
    }

    def _create_snapshot_entity_vertex(self, entity_event):

        stack_name = extract_field_value(entity_event, StackProps.STACK_NAME)
        stack_id = extract_field_value(entity_event, StackProps.ID)
        stack_state = extract_field_value(entity_event,
                                          StackProps.STACK_STATUS)
        timestamp = extract_field_value(entity_event, StackProps.CREATION_TIME)
        project_id = extract_field_value(entity_event, StackProps.PROJECT)

        return self._create_vertex(entity_event,
                                   stack_name,
                                   stack_id,
                                   stack_state,
                                   timestamp,
                                   project_id)

    def _create_update_entity_vertex(self, entity_event):

        volume_name = extract_field_value(entity_event, StackProps.STACK_NAME)
        volume_id = extract_field_value(entity_event,
                                        StackProps.STACK_IDENTITY)
        volume_state = extract_field_value(entity_event, StackProps.STATE)
        timestamp = entity_event.get(StackProps.CREATED_AT, None)
        project_id = entity_event.get(StackProps.TENANT_ID, None)

        return self._create_vertex(entity_event,
                                   volume_name,
                                   volume_id,
                                   volume_state,
                                   timestamp,
                                   project_id)

    def _create_vertex(self,
                       entity_event,
                       stack_name,
                       stack_id,
                       stack_state,
                       update_timestamp,
                       project_id):
        metadata = {
            VProps.NAME: stack_name,
            VProps.PROJECT_ID: project_id,
        }

        entity_key = self._create_entity_key(entity_event)

        vitrage_sample_timestamp = entity_event[DSProps.SAMPLE_DATE]

        return graph_utils.create_vertex(
            entity_key,
            vitrage_category=EntityCategory.RESOURCE,
            vitrage_type=HEAT_STACK_DATASOURCE,
            vitrage_sample_timestamp=vitrage_sample_timestamp,
            entity_id=stack_id,
            entity_state=stack_state,
            update_timestamp=update_timestamp,
            metadata=metadata)

    def _create_snapshot_neighbors(self, entity_event):
        return self._create_stack_neighbors(entity_event)

    def _create_update_neighbors(self, entity_event):
        return self._create_stack_neighbors(entity_event)

    def _create_entity_key(self, entity_event):

        is_update_event = tbase.is_update_event(entity_event)
        id_field_path = StackProps.STACK_IDENTITY\
            if is_update_event else StackProps.ID
        volume_id = extract_field_value(entity_event, id_field_path)

        key_fields = self._key_values(HEAT_STACK_DATASOURCE, volume_id)
        return build_key(key_fields)

    def _create_stack_neighbors(self, entity_event):
        neighbors = []

        for neighbor in entity_event[StackProps.RESOURCES]:
            neighbor_id = neighbor[StackProps.PHYSICAL_RESOURCE_ID]
            neighbor_datasource_type = \
                self.RESOURCE_TYPE[neighbor[StackProps.RESOURCES_TYPE]]
            neighbors.append(self._create_neighbor(entity_event,
                                                   neighbor_id,
                                                   neighbor_datasource_type,
                                                   EdgeLabel.COMPRISED,
                                                   is_entity_source=True))

        return neighbors

    def get_vitrage_type(self):
        return HEAT_STACK_DATASOURCE



# Copyright 2016 - Nokia Corporation
# Copyright 2014 OpenStack Foundation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# noinspection PyPackageRequirements

from vitrage.tests.functional import api


class HealthTest(api.FunctionalTest):

    def __init__(self, *args, **kwds):
        super(HealthTest, self).__init__(*args, **kwds)
        self.auth = 'keystone'

    def test_health(self):
        resp = self.app.get('/healthcheck')
        self.assertEqual('200 OK', resp.status)

    def test_health_json(self):
        resp = self.app.get('/healthcheck',
                            headers={'Accept': 'application/json'})
        self.assertEqual('200 OK', resp.status)
        self.assertDictEqual({'detailed': False, 'reasons': ['OK']},
                             resp.json, None)



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from io import StringIO
from lxml import etree
from oslo_log import log

from vitrage.datasources.nagios.properties import NagiosProperties

LOG = log.getLogger(__name__)


class NagiosParser(object):

    STATUS_TABLE_XPATH = '/html/body/table[@class="status"]'
    SERVICE_ROWS_XPATH = 'tr[position() > 1]'
    NAME_XPATH = 'table/tr/td[position()=1]/table/tr/td/a'

    def __init__(self):
        self.last_host_name = None

    def parse(self, html):
        try:
            tree = etree.parse(StringIO(html), etree.HTMLParser())
            status_tables = tree.xpath(self.STATUS_TABLE_XPATH)

            return self._parse_services(status_tables)

        except Exception:
            LOG.exception('Failed to get nagios services.')

    def _parse_services(self, status_tables):
        LOG.debug('Start parsing Nagios status')

        services = []
        for status_table in status_tables:
            service_rows = status_table.xpath(self.SERVICE_ROWS_XPATH)

            for service_row in service_rows:
                service = self._parse_service_row(service_row)

                if service:
                    LOG.debug('Appending service: %s', service)
                    services.append(service)
                else:
                    LOG.debug('service is None for service_row: %s',
                              service_row)

        LOG.debug('Done parsing Nagios status')

        return services

    def _parse_service_row(self, service_row):
        columns = list(service_row)

        # service lines have a fixed number of columns.
        # there are also two blank lines between different hosts,
        # so len(columns)==1 is also valid
        # TODO(ifat_afek): get column names by the header line
        if len(columns) == NagiosProperties.NUM_COLUMNS:
            return self._parse_service_columns(columns)

        elif len(columns) > NagiosProperties.NUM_COLUMNS:
            LOG.warning('Too many columns in nagios service row. '
                        'Found %d', len(columns))

        elif len(columns) > 1:
            LOG.warning('Missing columns in nagios service row. '
                        'Found only %d', len(columns))

        return None

    def _parse_service_columns(self, columns):
        return {
            NagiosProperties.RESOURCE_NAME:
                self._parse_host_name(columns[0], self.NAME_XPATH),
            NagiosProperties.SERVICE:
                self._get_column_data(columns[1], self.NAME_XPATH),
            NagiosProperties.STATUS: columns[2].text,
            NagiosProperties.LAST_CHECK: columns[3].text,
            NagiosProperties.DURATION: columns[4].text,
            NagiosProperties.ATTEMPT: columns[5].text,
            NagiosProperties.STATUS_INFO: columns[6].text
        }

    @staticmethod
    def _get_column_data(column, xpath):
        contents = column.xpath(xpath)

        if len(contents) == 1:
            return contents[0].text
        elif len(contents) > 1:
            LOG.warning('Multiple entries for nagios service column: %s',
                        contents.toString)
            return contents[0].text
        else:
            # len(contents) might be 0 for a host, since each host name appears
            # only once in the table
            return None

    def _parse_host_name(self, column, xpath):
        """Parse the host name and return it

        host name appears only once in the table, so keep using the same host
        name until found a new one

        :param column: the html column data
        :param xpath: the path of the host name in the column structure
        :return: the host name
        """
        host_name = self._get_column_data(column, xpath)

        if host_name:
            self.last_host_name = host_name
        else:
            host_name = self.last_host_name

        return host_name



# Copyright 2015 - Alcatel-Lucent
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


"""Methods for generating driver events

For each vitrage_type of entity, need to supply configuration files that
specify (a regex of) what can be returned, which will be used to generate
driver events

usage example:
    test_entity_spec_list = [
        {mg.DYNAMIC_INFO_FKEY: 'driver_inst_snapshot_dynamic.json',
         mg.STATIC_INFO_FKEY: 'driver_inst_snapshot_static.json',
         mg.MAPPING_KEY: [('vm1', 'host1'), ('vm2', 'host1'), ('vm3','host2')],
         mg.NAME_KEY: 'Instance (vm) generator',
         NUM_EVENTS_KEY: 10
         }
    ]
    spec_list = get_mock_generators(test_entity_spec_list)
    events = generate_random_events_list(spec_list)
    for e in events:
        print e
"""

import random

from vitrage.common.constants import DatasourceAction
from vitrage.common.constants import DatasourceProperties as DSProps
import vitrage.tests.mocks.trace_generator as tg
from vitrage.utils.datetime import utcnow


def generate_random_events_list(generator_spec_list):
    """Generates random events for the generators given.

     Each element in the list of generators includes a generator and
     number of events to generate for it's entities

     :param generator_spec_list: list of generators
     :type generator_spec_list: list

     :return: list of driver events
     :rtype list

    """

    data = []
    for spec in generator_spec_list:
        generator = spec[tg.GENERATOR]
        data += tg.generate_data_stream(generator.models, spec[tg.NUM_EVENTS])
    random.shuffle(data)
    return data


def generate_sequential_events_list(generator_spec_list):
    """Generates random events for the generators given.

     Each element in the list of generators includes a generator and
     number of events to generate for it's entities

     :param generator_spec_list: list of generators
     :type generator_spec_list: list

     :return: list of driver events
     :rtype list

    """

    data = []
    for spec in generator_spec_list:
        generator = spec[tg.GENERATOR]
        data += tg.generate_round_robin_data_stream(generator.models,
                                                    spec[tg.NUM_EVENTS])
    return data


def simple_instance_generators(host_num, vm_num,
                               snapshot_events=0, update_events=0,
                               use_nova_versioned_format=True,
                               snap_vals=None, update_vals=None):
    """A function for returning vm event generators.

    Returns generators for a given number of hosts and
    instances. Instances will be distributed across hosts in round-robin style.

    :param host_num: number of hosts
    :param vm_num: number of vms
    :param snapshot_events: number of snapshot events per instance
    :param update_events: number of update events per instance
    :param use_nova_versioned_format: use the format of Nova versioned
           notifications for the update events
    :param snap_vals: preset vals for ALL snapshot events
    :param update_vals: preset vals for ALL update events
    :return: generators for vm_num vms as specified
    """

    mapping = [('vm-{0}'.format(index), 'host-{0}'.format(index % host_num))
               for index in range(vm_num)
               ]

    test_entity_spec_list = []
    if snapshot_events:
        test_entity_spec_list.append(
            {tg.DYNAMIC_INFO_FKEY: tg.DRIVER_INST_SNAPSHOT_D,
             tg.STATIC_INFO_FKEY: tg.DRIVER_INST_SNAPSHOT_S,
             tg.EXTERNAL_INFO_KEY: snap_vals,
             tg.MAPPING_KEY: mapping,
             tg.NAME_KEY: 'Instance (vm) snapshot generator',
             tg.NUM_EVENTS: snapshot_events
             }
        )

    dynamic_info = tg.DRIVER_INST_UPDATE_VERSIONED_D \
        if use_nova_versioned_format else tg.DRIVER_INST_UPDATE_LEGACY_D

    if update_events:
        test_entity_spec_list.append(
            {tg.DYNAMIC_INFO_FKEY: dynamic_info,
             tg.STATIC_INFO_FKEY: None,
             tg.EXTERNAL_INFO_KEY: update_vals,
             tg.MAPPING_KEY: mapping,
             tg.NAME_KEY: 'Instance (vm) update generator',
             tg.NUM_EVENTS: update_events
             }
        )

    return tg.get_trace_generators(test_entity_spec_list)


def simple_host_generators(zone_num, host_num, snapshot_events=0,
                           snap_vals=None):
    """A function for returning vm event generators.

    Returns generators for a given number of hosts and
    instances. Instances will be distributed across hosts in round-robin style.

    :param zone_num: number of zones
    :param host_num: number of hosts
    :param snapshot_events: number of snapshot events per host
    :param snap_vals: preset vals for ALL snapshot events
    :return: generators for host_num hosts as specified
    """

    mapping = [('host-{0}'.format(index), 'zone-{0}'.format(index % zone_num))
               for index in range(host_num)
               ]

    test_entity_spec_list = []
    if snapshot_events:
        test_entity_spec_list.append(
            {tg.DYNAMIC_INFO_FKEY: tg.DRIVER_HOST_SNAPSHOT_D,
             tg.STATIC_INFO_FKEY: None,
             tg.EXTERNAL_INFO_KEY: snap_vals,
             tg.MAPPING_KEY: mapping,
             tg.NAME_KEY: 'Host snapshot generator',
             tg.NUM_EVENTS: snapshot_events
             }
        )
    return tg.get_trace_generators(test_entity_spec_list)


def simple_zone_generators(zone_num, host_num, snapshot_events=0,
                           snap_vals=None):
    """A function for returning zone event generators.

    Returns generators for a given number of hosts and
    zones. Hosts will be distributed across zones in round-robin style.

    :param zone_num: number of zones
    :param host_num: number of hosts
    :param snapshot_events: number of snapshot events per zone
    :param snap_vals: preset vals for ALL snapshot events
    :return: generators for zone_num zones as specified
    """

    mapping = [('host-{0}'.format(index), 'zone-{0}'.format(index % zone_num))
               for index in range(host_num)
               ]

    test_entity_spec_list = []
    if snapshot_events:
        test_entity_spec_list.append(
            {tg.DYNAMIC_INFO_FKEY: tg.DRIVER_ZONE_SNAPSHOT_D,
             tg.STATIC_INFO_FKEY: None,
             tg.EXTERNAL_INFO_KEY: snap_vals,
             tg.MAPPING_KEY: mapping,
             tg.NAME_KEY: 'Zone snapshot generator',
             tg.NUM_EVENTS: snapshot_events
             }
        )
    return tg.get_trace_generators(test_entity_spec_list)


def simple_trove_instance_generators(inst_num, snapshot_events=0,
                                     snap_vals=None):
    """A function for returning Trove instance generators.

    Returns generators for a given number of Trove instances.

    :param inst_num: number of instances
    :return: generators for inst_num instances as specified
    """

    mapping = [('tr-instance-{0}'.format(idx), 'vm-{0}'.format(idx))
               for idx in range(inst_num)]

    test_entity_spec_list = [
        {tg.DYNAMIC_INFO_FKEY: tg.DRIVER_TROVE_INSTANCE_SNAPSHOT_D,
         tg.STATIC_INFO_FKEY: None,
         tg.MAPPING_KEY: mapping,
         tg.EXTERNAL_INFO_KEY: snap_vals,
         tg.NAME_KEY: 'Database instance snapshot generator',
         tg.NUM_EVENTS: snapshot_events
         }
    ]

    return tg.get_trace_generators(test_entity_spec_list)


def simple_trove_cluster_generators(clust_num, inst_num, snapshot_events=0,
                                    snap_vals=None):
    """A function for returning Trove cluster generators.

    Returns generators for a given number of Trove clusters.

    :param clust_num: number of clusters
    :param inst_num: number of instances
    :return: generators for clust_num clusters as specified
    """

    mapping = [('tr-cluster-{0}'.format(idx % clust_num),
                'tr-inst-{0}'.format(idx))
               for idx in range(inst_num)]

    test_entity_spec_list = [
        {tg.DYNAMIC_INFO_FKEY: tg.DRIVER_TROVE_CLUSTER_SNAPSHOT_D,
         tg.STATIC_INFO_FKEY: None,
         tg.MAPPING_KEY: mapping,
         tg.EXTERNAL_INFO_KEY: snap_vals,
         tg.NAME_KEY: 'Database cluster snapshot generator',
         tg.NUM_EVENTS: snapshot_events
         }
    ]

    return tg.get_trace_generators(test_entity_spec_list)


def simple_volume_generators(volume_num, instance_num,
                             snapshot_events=0, update_events=0,
                             snap_vals=None, update_vals=None):
    """A function for returning vm event generators.

    Returns generators for a given number of volumes and
    instances. Instances will be distributed across hosts in round-robin style.

    :param update_vals:  number of values from update event
    :param update_events: number of events from update event
    :param volume_num: number of volumes
    :param instance_num: number of instances
    :param snapshot_events: number of snapshot events per host
    :param snap_vals: preset vals for ALL snapshot events
    :return: generators for volume_num volumes as specified
    """

    mapping = [('volume-{0}'.format(index % volume_num),
                'vm-{0}'.format(index))
               for index in range(instance_num)
               ]

    test_entity_spec_list = []
    if snapshot_events:
        test_entity_spec_list.append(
            {tg.DYNAMIC_INFO_FKEY: tg.DRIVER_VOLUME_SNAPSHOT_D,
             tg.STATIC_INFO_FKEY: None,
             tg.EXTERNAL_INFO_KEY: snap_vals,
             tg.MAPPING_KEY: mapping,
             tg.NAME_KEY: 'Volume snapshot generator',
             tg.NUM_EVENTS: snapshot_events
             }
        )
    if update_events:
        test_entity_spec_list.append(
            {tg.DYNAMIC_INFO_FKEY: tg.DRIVER_VOLUME_UPDATE_D,
             tg.STATIC_INFO_FKEY: None,
             tg.EXTERNAL_INFO_KEY: update_vals,
             tg.MAPPING_KEY: mapping,
             tg.NAME_KEY: 'Volume update generator',
             tg.NUM_EVENTS: update_events
             }
        )
    return tg.get_trace_generators(test_entity_spec_list)


def simple_stack_generators(stack_num, instance_and_volume_num,
                            snapshot_events=0, update_events=0,
                            snap_vals=None, update_vals=None):
    """A function for returning vm event generators.

    Returns generators for a given number of stacks, instances and
    volumes. Instances and Volumes will be distributed across stacks in
    round-robin style.

    :param update_vals:  number of values from update event
    :param update_events: number of events from update event
    :param stack_num: number of stacks
    :param instance_and_volume_num: number of volumes and instances
    :param snapshot_events: number of snapshot events per host
    :param snap_vals: preset vals for ALL snapshot events
    :return: generators for volume_num volumes as specified
    """

    mapping = [('stack-{0}'.format(index % stack_num),
                'stack-vm-{0}'.format(index),
                'stack-volume-{0}')
               for index in range(instance_and_volume_num)
               ]

    test_entity_spec_list = []
    if snapshot_events:
        test_entity_spec_list.append(
            {tg.DYNAMIC_INFO_FKEY: tg.DRIVER_STACK_SNAPSHOT_D,
             tg.STATIC_INFO_FKEY: None,
             tg.EXTERNAL_INFO_KEY: snap_vals,
             tg.MAPPING_KEY: mapping,
             tg.NAME_KEY: 'Stack snapshot generator',
             tg.NUM_EVENTS: snapshot_events
             }
        )
    if update_events:
        test_entity_spec_list.append(
            {tg.DYNAMIC_INFO_FKEY: tg.DRIVER_STACK_UPDATE_D,
             tg.STATIC_INFO_FKEY: None,
             tg.EXTERNAL_INFO_KEY: update_vals,
             tg.MAPPING_KEY: mapping,
             tg.NAME_KEY: 'Stack update generator',
             tg.NUM_EVENTS: update_events
             }
        )
    return tg.get_trace_generators(test_entity_spec_list)


def simple_consistency_generators(consistency_num, update_events=0,
                                  snap_vals=None, update_vals=None):
    """A function for returning consistency event generators.

    Returns generators for a given number of consistency events.
    Instances will be distributed across hosts in round-robin style.

    :param update_vals:  number of values from update event
    :param update_events: number of events from update event
    :param consistency_num: number of consistency events
    :param snap_vals: preset vals for ALL snapshot events
    :return: generators for consistency_num consistency events as specified
    """

    test_entity_spec_list = []
    if update_events:
        test_entity_spec_list.append(
            {tg.DYNAMIC_INFO_FKEY: tg.DRIVER_CONSISTENCY_UPDATE_D,
             tg.STATIC_INFO_FKEY: None,
             tg.EXTERNAL_INFO_KEY: update_vals,
             tg.MAPPING_KEY: consistency_num,
             tg.NAME_KEY: 'Consistency update generator',
             tg.NUM_EVENTS: update_events
             }
        )
    return tg.get_trace_generators(test_entity_spec_list)


def simple_static_generators(switch_num=2, host_num=10,
                             snapshot_events=0, snap_vals=None,
                             update_events=0, update_vals=None):
    """A function for returning static datasource events generators.

    Returns generators for a given number of routers, switches and hosts.
    Hosts will be distributed across switches in round-robin style.
    Switches are interconnected in a line.

    :param switch_num: number of zones
    :param host_num: number of hosts
    :param snapshot_events: number of snapshot events per zone
    :param snap_vals: preset values for ALL snapshot events
    :param update_events: number of values from update event
    :param update_vals: preset values for update event

    :return: generators for static datasource events
    """

    # TODO(yujunz) mock routers which connects all switches
    mapping = [(host_index, host_index % switch_num)
               for host_index in range(host_num)]

    test_entity_spec_list = []
    if snapshot_events > 0:
        if snap_vals is None:
            snap_vals = {}
        snap_vals.update({
            DSProps.DATASOURCE_ACTION: DatasourceAction.SNAPSHOT,
            DSProps.SAMPLE_DATE: str(utcnow())})
        test_entity_spec_list.append(
            {tg.DYNAMIC_INFO_FKEY: tg.DRIVER_STATIC_SNAPSHOT_D,
             tg.STATIC_INFO_FKEY: tg.DRIVER_STATIC_SNAPSHOT_S,
             tg.EXTERNAL_INFO_KEY: snap_vals,
             tg.MAPPING_KEY: mapping,
             tg.NAME_KEY: 'Static snapshot generator',
             tg.NUM_EVENTS: snapshot_events
             }
        )
    if update_events > 0:
        if update_vals is None:
            update_vals = {}
        update_vals.update({
            DSProps.DATASOURCE_ACTION: DatasourceAction.UPDATE,
            DSProps.SAMPLE_DATE: str(utcnow())})
        test_entity_spec_list.append(
            {tg.DYNAMIC_INFO_FKEY: tg.DRIVER_STATIC_SNAPSHOT_D,
             tg.STATIC_INFO_FKEY: None,
             tg.EXTERNAL_INFO_KEY: update_vals,
             tg.MAPPING_KEY: mapping,
             tg.NAME_KEY: 'Static update generator',
             tg.NUM_EVENTS: update_events
             }
        )
    return tg.get_trace_generators(test_entity_spec_list)


def simple_nagios_alarm_generators(host_num,
                                   events_num=0,
                                   snap_vals=None):
    """A function for returning Nagios alarm event generators.

    Returns generators for a given number of Nagios alarms.

    :param host_num: number of hosts
    :param events_num: number of snapshot alarms per hosts
    :param snap_vals: preset vals for ALL snapshot events
    :return: generators for zone_num zones as specified
    """

    hosts = ['host-{0}'.format(index) for index in range(host_num)]

    test_entity_spec_list = []
    if events_num:
        test_entity_spec_list.append({
            tg.DYNAMIC_INFO_FKEY: tg.DRIVER_NAGIOS_SNAPSHOT_D,
            tg.STATIC_INFO_FKEY: None,
            tg.EXTERNAL_INFO_KEY: snap_vals,
            tg.MAPPING_KEY: hosts,
            tg.NAME_KEY: 'Nagios alarm generator (alarm on)',
            tg.NUM_EVENTS: max(events_num - len(hosts), 0)
        })
        test_entity_spec_list.append({
            tg.DYNAMIC_INFO_FKEY: tg.DRIVER_NAGIOS_SNAPSHOT_D,
            tg.STATIC_INFO_FKEY: tg.DRIVER_NAGIOS_SNAPSHOT_S,
            tg.EXTERNAL_INFO_KEY: snap_vals,
            tg.MAPPING_KEY: hosts,
            tg.NAME_KEY: 'Nagios alarm generator (alarm off)',
            tg.NUM_EVENTS: len(hosts)
        })

    return tg.get_trace_generators(test_entity_spec_list)


def simple_zabbix_alarm_generators(host_num,
                                   events_num=0,
                                   snap_vals=None):
    """A function for returning Zabbix alarm event generators.

    Returns generators for a given number of Zabbix alarms.

    :param host_num: number of hosts
    :param events_num: number of snapshot alarms per hosts
    :param snap_vals: preset vals for ALL snapshot events
    :return: generators for zone_num zones as specified
    """

    hosts = ['host-{0}'.format(index) for index in range(host_num)]

    test_entity_spec_list = []
    if events_num:
        test_entity_spec_list.append({
            tg.DYNAMIC_INFO_FKEY: tg.DRIVER_ZABBIX_SNAPSHOT_D,
            tg.STATIC_INFO_FKEY: None,
            tg.EXTERNAL_INFO_KEY: snap_vals,
            tg.MAPPING_KEY: hosts,
            tg.NAME_KEY: 'Zabbix alarm generator (alarm on)',
            tg.NUM_EVENTS: max(events_num - len(hosts), 0)
        })
        test_entity_spec_list.append({
            tg.DYNAMIC_INFO_FKEY: tg.DRIVER_ZABBIX_SNAPSHOT_D,
            tg.STATIC_INFO_FKEY: None,
            tg.EXTERNAL_INFO_KEY: snap_vals,
            tg.MAPPING_KEY: hosts,
            tg.NAME_KEY: 'Zabbix alarm generator (alarm off)',
            tg.NUM_EVENTS: len(hosts)
        })

    return tg.get_trace_generators(test_entity_spec_list)


def simple_doctor_alarm_generators(update_vals=None):
    """A function for returning Doctor alarm event generators.

    Returns generators for a given number of Doctor alarms.

    :param update_vals: preset values for ALL update events
    :return: generators for alarms as specified
    """

    test_entity_spec_list = [({
        tg.DYNAMIC_INFO_FKEY: tg.DRIVER_DOCTOR_UPDATE_D,
        tg.STATIC_INFO_FKEY: None,
        tg.EXTERNAL_INFO_KEY: update_vals,
        tg.MAPPING_KEY: None,
        tg.NAME_KEY: 'Doctor alarm generator',
        tg.NUM_EVENTS: 1
    })]

    return tg.get_trace_generators(test_entity_spec_list)


def simple_collectd_alarm_generators(update_vals=None):
    """A function for returning Collectd alarm event generators.

    Returns generators for a given number of Collectd alarms.

    :param update_vals: preset values for ALL update events
    :return: generators for alarms as specified
    """

    test_entity_spec_list = [({
        tg.DYNAMIC_INFO_FKEY: tg.DRIVER_COLLECTD_UPDATE_D,
        tg.STATIC_INFO_FKEY: None,
        tg.EXTERNAL_INFO_KEY: update_vals,
        tg.MAPPING_KEY: None,
        tg.NAME_KEY: 'Collectd alarm generator',
        tg.NUM_EVENTS: 1
    })]

    return tg.get_trace_generators(test_entity_spec_list)


def simple_aodh_alarm_notification_generators(alarm_num,
                                              update_events=0,
                                              update_vals=None):
    """A function for returning aodh alarm event generators.

    Returns generators for a given number of Aodh alarms.

    :param alarm_num: number of alarms
    :param update_events: number of update alarms
    :param update_vals: preset vals for ALL update events
    :return: generators for alarm_num zones as specified

    Returns generators for a given number of alarms and
    instances.
    """

    alarms = ['alarm-{0}'.format(index) for index in range(alarm_num)]

    test_entity_spec_list = [
        {tg.DYNAMIC_INFO_FKEY: tg.DRIVER_AODH_UPDATE_D,
         tg.STATIC_INFO_FKEY: None,
         tg.MAPPING_KEY: alarms,
         tg.EXTERNAL_INFO_KEY: update_vals,
         tg.NAME_KEY: 'Aodh update generator',
         tg.NUM_EVENTS: update_events
         }
    ]

    return tg.get_trace_generators(test_entity_spec_list)


def simple_prometheus_alarm_generators(update_vals=None):
    """A function for returning Prometheus alarm event generators.

    Returns generators for a given number of Prometheus alarms.

    :param update_vals: preset values for ALL update events
    :return: generators for alarms as specified
    """

    test_entity_spec_list = [({
        tg.DYNAMIC_INFO_FKEY: tg.DRIVER_PROMETHEUS_UPDATE_D,
        tg.STATIC_INFO_FKEY: None,
        tg.EXTERNAL_INFO_KEY: update_vals,
        tg.MAPPING_KEY: None,
        tg.NAME_KEY: 'Prometheus alarm generator',
        tg.NUM_EVENTS: 1
    })]

    return tg.get_trace_generators(test_entity_spec_list)


def simple_k8s_nodes_generators(nodes_num, snapshot_events=0):
    mapping = ['vm-{0}'.format(index) for index in range(nodes_num)]

    test_entity_spec_list = []
    if snapshot_events:
        test_entity_spec_list.append(
            {tg.DYNAMIC_INFO_FKEY: tg.DRIVER_KUBE_SNAPSHOT_D,
             tg.STATIC_INFO_FKEY: tg.DRIVER_INST_SNAPSHOT_S,
             tg.MAPPING_KEY: mapping,
             tg.NAME_KEY: 'Nodes snapshot generator',
             tg.NUM_EVENTS: snapshot_events
             }
        )
    return tg.get_trace_generators(test_entity_spec_list)


def simple_kapacitor_alarm_generators(update_vals=None):
    """A function for returning Kapacitor alarm event generators.

    Returns generators for a given number of Kapacitor alarms.

    :param update_vals: preset values for ALL update events
    :return: generators for alarms as specified
    """

    test_entity_spec_list = [({
        tg.DYNAMIC_INFO_FKEY: tg.DRIVER_KAPACITOR_UPDATE_D,
        tg.STATIC_INFO_FKEY: None,
        tg.EXTERNAL_INFO_KEY: update_vals,
        tg.MAPPING_KEY: None,
        tg.NAME_KEY: 'Kapacitor alarm generator',
        tg.NUM_EVENTS: 1
    })]

    return tg.get_trace_generators(test_entity_spec_list)



# Copyright 2019 - Viettel
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.


from oslo_log import log as logging

from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.common.constants import EdgeLabel
from vitrage.common.constants import EntityCategory
from vitrage.common.constants import VertexProperties as VProps
from vitrage.datasources.alarm_transformer_base import \
    AlarmTransformerBase
from vitrage.datasources.kapacitor import KAPACITOR_DATASOURCE
from vitrage.datasources.kapacitor.properties import KapacitorProperties \
    as KProps
from vitrage.datasources.kapacitor.properties import KapacitorState
from vitrage.datasources import transformer_base as tbase
import vitrage.graph.utils as graph_utils

LOG = logging.getLogger(__name__)


class KapacitorTransformer(AlarmTransformerBase):

    def _create_snapshot_entity_vertex(self, entity_event):
        return self._create_vertex(entity_event)

    def _create_update_entity_vertex(self, entity_event):
        return self._create_vertex(entity_event)

    def _create_snapshot_neighbors(self, entity_event):
        return self._create_kapacitor_neighbors(entity_event)

    def _create_update_neighbors(self, entity_event):
        return self._create_kapacitor_neighbors(entity_event)

    def _create_entity_key(self, entity_event):
        """the unique key of this entity"""
        entity_type = entity_event[DSProps.ENTITY_TYPE]
        alarm_id = entity_event[KProps.ID]
        resource_name = entity_event[KProps.RESOURCE_NAME]
        return tbase.build_key((EntityCategory.ALARM,
                                entity_type,
                                resource_name,
                                alarm_id))

    def get_vitrage_type(self):
        return KAPACITOR_DATASOURCE

    def _create_vertex(self, entity_event):
        update_timestamp = entity_event[KProps.TIME]

        vitrage_sample_timestamp = entity_event[DSProps.SAMPLE_DATE]

        update_timestamp = \
            self._format_update_timestamp(update_timestamp,
                                          vitrage_sample_timestamp)
        metadata = {
            VProps.NAME: entity_event[KProps.MESSAGE],
            VProps.SEVERITY: entity_event[KProps.PRIORITY],
            VProps.RAWTEXT: entity_event[KProps.DETAILS],
            VProps.RESOURCE_NAME: entity_event[KProps.RESOURCE_NAME],
        }

        return graph_utils.create_vertex(
            self._create_entity_key(entity_event),
            vitrage_category=EntityCategory.ALARM,
            vitrage_type=entity_event[DSProps.ENTITY_TYPE],
            vitrage_sample_timestamp=vitrage_sample_timestamp,
            update_timestamp=update_timestamp,
            entity_state=self._get_alarm_state(entity_event),
            metadata=metadata)

    def _ok_status(self, entity_event):
        return entity_event[KProps.PRIORITY] == KapacitorState.OK

    def _create_kapacitor_neighbors(self, entity_event):
        graph_neighbors = entity_event.get(self.QUERY_RESULT, [])
        return [self._create_neighbor(
                entity_event,
                graph_neighbor[VProps.ID],
                graph_neighbor[VProps.VITRAGE_TYPE],
                EdgeLabel.ON,
                neighbor_category=EntityCategory.RESOURCE)
                for graph_neighbor in graph_neighbors]

    @staticmethod
    def get_enrich_query(event):
        resource_type = event.get(KProps.RESOURCE_TYPE)
        resource_name = event.get(KProps.RESOURCE_NAME)

        if resource_type and resource_name:
            return {VProps.NAME: resource_name,
                    VProps.VITRAGE_TYPE: resource_type}

        return None



# Copyright 2015 - Alcatel-Lucent
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import logging
import os

from oslo_config import cfg
from oslo_config import fixture as config_fixture
from oslo_utils import timeutils
# noinspection PyPackageRequirements
from oslotest import base
import sys

from testtools import matchers
from testtools.matchers import HasLength

from vitrage.common import config

CONF = cfg.CONF
IsEmpty = lambda: HasLength(0)


class BaseTest(base.BaseTestCase):
    """Test case base class for all unit tests."""

    def conf_reregister_opts(self, opts, group=None):
        self.conf.reset()
        if group in self.conf:
            for opt in opts:
                self.conf.unregister_opt(opt, group=group)
        self.conf.register_opts(opts, group=group)

        def unregister_opts():
            self.conf.reset()
            for opt in opts:
                self.conf.unregister_opt(opt, group=group)

        self.addCleanup(unregister_opts)

    def setUp(self):
        super(BaseTest, self).setUp()
        self.cfg_fixture = self.useFixture(
            config_fixture.Config(CONF))
        config.parse_config([])
        logging.disable(logging.CRITICAL)
        self.conf = self.cfg_fixture.conf

    def config(self, **kw):
        self.cfg_fixture.config(**kw)

    def assert_list_equal(self, l1, l2, message=None):
        if tuple(sys.version_info)[0:2] < (2, 7):
            # for python 2.6 compatibility
            self.assertEqual(l1, l2, message)
        else:
            super(BaseTest, self).assertListEqual(l1, l2, message)

    def assert_dict_equal(self, d1, d2, message=None):
        if tuple(sys.version_info)[0:2] < (2, 7):
            # for python 2.6 compatibility
            self.assertEqual(d1, d2)
        else:
            super(BaseTest, self).assertDictEqual(d1, d2, message)

    def assert_timestamp_equal(self, first, second, msg=None):
        """Checks that two timestamps are equals.

        This relies on assertAlmostEqual to avoid rounding problem, and only
        checks up the first microsecond values.

        """
        return self.assertAlmostEqual(timeutils.delta_seconds(first, second),
                                      0.0,
                                      places=5, msg=msg)

    def assert_is_empty(self, obj):
        try:
            if len(obj) != 0:
                self.fail("%s is not empty" % type(obj))
        except (TypeError, AttributeError):
            self.fail("%s doesn't have length" % type(obj))

    def assert_is_not_empty(self, obj):
        try:
            if len(obj) == 0:
                self.fail("%s is empty" % type(obj))
        except (TypeError, AttributeError):
            self.fail("%s doesn't have length" % type(obj))

    def assert_graph_equal(self, g1, g2):
        """Checks that two graphs are equals.

        This relies on assert_dict_equal when comparing the nodes and the
        edges of each graph.
        """
        g1_nodes = g1._g.nodes
        g1_edges = g1._g.adj
        g2_nodes = g2._g.nodes
        g2_edges = g2._g.adj
        self.assertEqual(g1.num_vertices(), g2.num_vertices(),
                         "Two graphs have different amount of nodes")
        self.assertEqual(g1.num_edges(), g2.num_edges(),
                         "Two graphs have different amount of edges")
        for n_id in g1_nodes:
            self.assert_dict_equal(g1_nodes.get(n_id),
                                   g2_nodes.get(n_id),
                                   "Nodes of each graph are not equal")

        for e_source_id in g1_edges:
            self.assert_dict_equal(dict(g1_edges.get(e_source_id)),
                                   dict(g2_edges.get(e_source_id)),
                                   "Edges of each graph are not equal")

    def assert_starts_with(self, expected_prefix, observed_str, msg=None):
        self.assertThat(observed_str,
                        matchers.StartsWith(expected_prefix), msg)

    @staticmethod
    def path_get(project_file=None):
        root = os.path.abspath(os.path.join(os.path.dirname(__file__),
                                            '..',
                                            '..',
                                            )
                               )
        if project_file:
            return os.path.join(root, project_file)
        else:
            return root



# Copyright 2016 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import copy

from copy import deepcopy
from oslo_log import log
from oslo_utils import importutils

from vitrage.common.constants import DatasourceAction as AType
from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.common.constants import VertexProperties as VProps
from vitrage.evaluator.actions.base import ActionMode
from vitrage.evaluator.actions.base import ActionType
from vitrage.evaluator.actions.evaluator_event_transformer \
    import VITRAGE_DATASOURCE
from vitrage.evaluator.actions.notifier import EvaluatorNotifier
from vitrage.evaluator.actions.recipes.action_steps import ADD_EDGE
from vitrage.evaluator.actions.recipes.action_steps import ADD_VERTEX
from vitrage.evaluator.actions.recipes.action_steps import EXECUTE_EXTERNAL
from vitrage.evaluator.actions.recipes.action_steps import EXECUTION_ENGINE
from vitrage.evaluator.actions.recipes.action_steps import REMOVE_EDGE
from vitrage.evaluator.actions.recipes.action_steps import REMOVE_VERTEX
from vitrage.evaluator.actions.recipes.action_steps import UPDATE_VERTEX
from vitrage.evaluator.actions.recipes.add_causal_relationship import \
    AddCausalRelationship
from vitrage.evaluator.actions.recipes.base import EVALUATOR_EVENT_TYPE
from vitrage.evaluator.actions.recipes.execute_mistral import ExecuteMistral
from vitrage.evaluator.actions.recipes.mark_down import MarkDown
from vitrage.evaluator.actions.recipes.raise_alarm import RaiseAlarm
from vitrage.evaluator.actions.recipes.set_state import SetState
from vitrage.utils import datetime as datetime_utils

LOG = log.getLogger(__name__)

EVALUATOR_EVENT = 'evaluator.event'
TARGET = 'target'
SOURCE = 'source'


class ActionExecutor(object):

    def __init__(self, actions_callback):

        self.actions_callback = actions_callback
        self.notifier = EvaluatorNotifier()
        self.action_recipes = ActionExecutor._register_action_recipes()

        self.action_step_defs = {
            ADD_VERTEX: self._add_vertex,
            REMOVE_VERTEX: self._remove_vertex,
            UPDATE_VERTEX: self._update_vertex,
            ADD_EDGE: self._add_edge,
            REMOVE_EDGE: self._remove_edge,
            EXECUTE_EXTERNAL: self._execute_external,
        }

    def execute(self, actions):
        if not actions:
            return

        events = []
        for action in actions:
            LOG.info('Action: %s', self._action_str(action))
            events.extend(self._execute(action.specs, action.mode))
        self.actions_callback(EVALUATOR_EVENT, events)

    def _execute(self, action_spec, action_mode):

        action_recipe = self.action_recipes[action_spec.type]
        if action_mode == ActionMode.DO:
            steps = action_recipe.get_do_recipe(action_spec)
        else:
            steps = action_recipe.get_undo_recipe(action_spec)

        events = []
        for step in steps:
            event = self.action_step_defs[step.type](step.params)
            if event:
                events.append(event)
        return events

    def _add_vertex(self, params):

        event = copy.deepcopy(params)
        ActionExecutor._add_default_properties(event)
        event[EVALUATOR_EVENT_TYPE] = ADD_VERTEX

        return event

    def _update_vertex(self, params):

        event = copy.deepcopy(params)
        ActionExecutor._add_default_properties(event)
        event[EVALUATOR_EVENT_TYPE] = UPDATE_VERTEX

        return event

    def _remove_vertex(self, params):
        event = copy.deepcopy(params)
        ActionExecutor._add_default_properties(event)
        event[EVALUATOR_EVENT_TYPE] = REMOVE_VERTEX

        return event

    def _add_edge(self, params):

        event = copy.deepcopy(params)
        ActionExecutor._add_default_properties(event)
        event[EVALUATOR_EVENT_TYPE] = ADD_EDGE

        return event

    def _remove_edge(self, params):

        event = copy.deepcopy(params)
        ActionExecutor._add_default_properties(event)
        event[EVALUATOR_EVENT_TYPE] = REMOVE_EDGE

        return event

    def _execute_external(self, params):

        # Send a notification to the external engine
        execution_engine = params[EXECUTION_ENGINE]
        payload = deepcopy(params)
        del payload[EXECUTION_ENGINE]

        LOG.debug('Notifying external engine %s. Properties: %s',
                  execution_engine,
                  payload)
        self.notifier.notify(execution_engine, payload)

    @staticmethod
    def _add_default_properties(event):

        event[DSProps.DATASOURCE_ACTION] = AType.UPDATE
        event[DSProps.ENTITY_TYPE] = VITRAGE_DATASOURCE
        event[VProps.UPDATE_TIMESTAMP] = str(datetime_utils.utcnow(False))
        event[VProps.VITRAGE_SAMPLE_TIMESTAMP] = datetime_utils.format_utcnow()

    @staticmethod
    def _register_action_recipes():

        # noinspection PyDictCreation
        recipes = {}

        recipes[ActionType.SET_STATE] = importutils.import_object(
            "%s.%s" % (SetState.__module__, SetState.__name__))

        recipes[ActionType.RAISE_ALARM] = importutils.import_object(
            "%s.%s" % (RaiseAlarm.__module__, RaiseAlarm.__name__))

        recipes[ActionType.ADD_CAUSAL_RELATIONSHIP] = \
            importutils.import_object(
            "%s.%s" % (AddCausalRelationship.__module__,
                       AddCausalRelationship.__name__))

        recipes[ActionType.MARK_DOWN] = importutils.import_object(
            "%s.%s" % (MarkDown.__module__, MarkDown.__name__))

        recipes[ActionType.EXECUTE_MISTRAL] = importutils.import_object(
            "%s.%s" % (ExecuteMistral.__module__, ExecuteMistral.__name__))

        return recipes

    @staticmethod
    def _action_str(action):
        s = action.specs.targets.get(SOURCE, {}).get(VProps.VITRAGE_ID, '')
        t = action.specs.targets.get(TARGET, {}).get(VProps.VITRAGE_ID, '')
        return '%s %s \'%s\' targets (%s,%s)' % (action.mode.upper(),
                                                 action.specs.type,
                                                 action.action_id, s, t)



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from datetime import datetime
from oslo_config import cfg

from vitrage.common.constants import DatasourceOpts as DSOpts
from vitrage.common.constants import DatasourceProperties as DSProps
from vitrage.common.constants import EventProperties as EventProps
from vitrage.common.constants import UpdateMethod
from vitrage.datasources.doctor import DOCTOR_DATASOURCE
from vitrage.datasources.doctor.properties import DoctorDetails
from vitrage.datasources.doctor.properties import DoctorProperties \
    as DoctorProps
from vitrage.datasources.doctor.properties import DoctorStatus
from vitrage.datasources.doctor.transformer import DoctorTransformer
from vitrage.datasources.nova.host import NOVA_HOST_DATASOURCE
from vitrage.datasources.nova.host.transformer import HostTransformer
from vitrage.datasources.transformer_base import TransformerBase
from vitrage.tests.mocks import mock_transformer
from vitrage.tests.unit.datasources.test_alarm_transformer_base import \
    BaseAlarmTransformerTest


# noinspection PyProtectedMember
class DoctorTransformerTest(BaseAlarmTransformerTest):

    OPTS = [
        cfg.StrOpt(DSOpts.UPDATE_METHOD,
                   default=UpdateMethod.PUSH),
    ]

    def setUp(self):
        super(DoctorTransformerTest, self).setUp()
        self.conf_reregister_opts(self.OPTS, group=DOCTOR_DATASOURCE)
        self.transformers = {}
        self.transformers[DOCTOR_DATASOURCE] = \
            DoctorTransformer(self.transformers)
        self.transformers[NOVA_HOST_DATASOURCE] = \
            HostTransformer(self.transformers)

    def test_create_update_entity_vertex(self):
        # Test setup
        time1 = datetime.now().isoformat()
        host1 = 'host1'
        event = self._generate_event(time1, host1, DoctorStatus.DOWN)
        self.assertIsNotNone(event)

        # Test action
        transformer = self.transformers[DOCTOR_DATASOURCE]
        wrapper = transformer.transform(event)

        # Test assertions
        self._validate_vertex_props(wrapper.vertex, event)
        entity_key1 = transformer._create_entity_key(event)
        entity_uuid1 = transformer.uuid_from_deprecated_vitrage_id(entity_key1)
        # Validate the neighbors: only one valid host neighbor
        self._validate_host_neighbor(wrapper, entity_uuid1, host1)

        # Validate the expected action on the graph - update or delete
        self._validate_graph_action(wrapper)

        # Create an event with status 'UP'
        time2 = datetime.now().isoformat()
        host2 = 'host2'
        event = self._generate_event(time2, host2, DoctorStatus.UP)
        self.assertIsNotNone(event)

        # Test action
        # after transform vitrage uuid will be deleted from uuid cache
        entity_key2 = transformer._create_entity_key(event)
        entity_uuid2 = \
            TransformerBase.uuid_from_deprecated_vitrage_id(entity_key2)

        transformer = self.transformers[DOCTOR_DATASOURCE]
        wrapper = transformer.transform(event)

        # Test assertions
        self._validate_vertex_props(wrapper.vertex, event)
        self._validate_host_neighbor(wrapper, entity_uuid2, host2)
        self._validate_graph_action(wrapper)

    def _validate_vertex_props(self, vertex, event):
        self._validate_alarm_vertex_props(vertex,
                                          event[EventProps.TYPE],
                                          DOCTOR_DATASOURCE,
                                          event[DSProps.SAMPLE_DATE])

    @staticmethod
    def _generate_event(time, hostname, status):
        details = {}
        if hostname:
            details[DoctorDetails.HOSTNAME] = hostname
        if status:
            details[DoctorDetails.STATUS] = status

        update_vals = {EventProps.DETAILS: details}
        if time:
            update_vals[EventProps.TIME] = time
            update_vals[DoctorProps.UPDATE_TIME] = time

        generators = mock_transformer.simple_doctor_alarm_generators(
            update_vals=update_vals)

        return mock_transformer.generate_random_events_list(generators)[0]

    def _is_erroneous(self, vertex):
        return vertex[DoctorDetails.STATUS] == DoctorStatus.DOWN



# Copyright 2017 - Nokia
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_log import log

from vitrage.common.constants import EntityCategory
from vitrage.evaluator.template_fields import TemplateFields
from vitrage.evaluator.template_validation.content.base import \
    ActionValidator
from vitrage.evaluator.template_validation.content.base import \
    get_content_correct_result
from vitrage.evaluator.template_validation.content.base import \
    get_content_fault_result
from vitrage.evaluator.template_validation.content.base import \
    validate_template_id
from vitrage.evaluator.template_validation.status_messages import status_msgs


LOG = log.getLogger(__name__)


class AddCausalRelationshipValidator(ActionValidator):

    @staticmethod
    def validate(action, definitions_index):
        if TemplateFields.ACTION_TARGET not in action:
            LOG.error('%s status code: %s' % (status_msgs[124], 124))
            return get_content_fault_result(124)

        action_target = action[TemplateFields.ACTION_TARGET]

        for key in [TemplateFields.TARGET, TemplateFields.SOURCE]:
            if key not in action_target:
                LOG.error('%s status code: %s' % (status_msgs[130], 130))
                return get_content_fault_result(130)

            template_id = action_target[key]
            result = validate_template_id(definitions_index, template_id)

            if not result.is_valid_config:
                return result

            entity = definitions_index[template_id]
            result = AddCausalRelationshipValidator._validate_entity_category(
                entity,
                EntityCategory.ALARM)
            if not result.is_valid_config:
                return result

        return get_content_correct_result()

    @staticmethod
    def _validate_entity_category(entity_to_check, category):

        if TemplateFields.CATEGORY not in entity_to_check \
                or entity_to_check[TemplateFields.CATEGORY] != category:
            msg = status_msgs[132] + ' expect %s to be %s' \
                                     % (entity_to_check, category)
            LOG.error('%s status code: %s' % (msg, 132))
            return get_content_fault_result(132, msg)

        return get_content_correct_result()
